from ply import yacc
import os
from lexer import tokens, pretty_error

from kurt.blockspecs import find_block, blocks_by_cmd
from kurt.scripts import Block, Script, SpriteRef
from kurt import Symbol, Color


import htmlcolor
htmlcolor_parser = htmlcolor.Parser()


DEBUG = False



class Insert(object):
    NUMBER = "NUMBER"      # ( )
    VARIABLE = "VARIABLE"
    REPORTER = "REPORTER"
    SPECIAL = "SPECIAL"
       
    STRING = "STRING"                    # [ ]
    STRING_DROPDOWN = "STRING_DROPDOWN"  # [ v]
    
    BOOL = "BOOL"  # < >
    
    COLOR = "COLOR" # [#fff]
    # COLOR is only used in `insert_kind`

    def __init__(self, kind, value):
        self.kind = kind
        self.value = value
    
    def __repr__(self):
        return "Insert(%r, %r)" % (self.kind, self.value)


class BlockError(Exception):
    pass


  
special_variables = {
    'all': Symbol('all'),
    'any': Symbol('any'),
    'last': Symbol('last'),
}

special_strings = {
    'mouse-pointer': Symbol('mouse'),
}

math_functions = ['sqrt', 'abs', 'sin', 'cos', 'tan', 'asin', 'acos', 'atan', 
                  'ln', 'log', 'e ^', '10 ^']

insert_kind = {
    "%b": [Insert.BOOL],
    "%n": [Insert.NUMBER, Insert.VARIABLE],
    "%d": [Insert.NUMBER, Insert.VARIABLE],
    "%s": [Insert.STRING, Insert.VARIABLE],
    
    "%c": [Insert.COLOR], # Color
    "%C": [Insert.COLOR],
    
    "%m": [Insert.STRING_DROPDOWN, Insert.VARIABLE, Insert.SPECIAL], # Sprite
                                                    # Symbol('mouse')
    "%a": [Insert.STRING_DROPDOWN], # attribute of sprite
    "%e": [Insert.STRING_DROPDOWN, Insert.VARIABLE], # broadcast message
    "%k": [Insert.STRING_DROPDOWN], # key
    
    "%v": [Insert.STRING_DROPDOWN], # var doesn't *accept* VARIABLE Insert!
    "%L": [Insert.STRING_DROPDOWN], # List name
    "%i": [Insert.NUMBER, Insert.SPECIAL, Insert.VARIABLE], # Item
                                   # Symbol('last'), Symbol('any')
    "%y": [Insert.NUMBER, Insert.SPECIAL, Insert.VARIABLE], # Item (for delete)
                                   # Symbol('last'), Symbol('all')
    
    "%f": [Insert.STRING_DROPDOWN], # Math function
           # one of `math_functions`
    "%l": [Insert.STRING_DROPDOWN, Insert.VARIABLE], # Costume name
    "%g": [Insert.STRING_DROPDOWN], # graphic effect
    
    "%S": [Insert.STRING_DROPDOWN, Insert.VARIABLE], # sound
    "%D": [Insert.NUMBER, Insert.VARIABLE], # drum
    "%N": [Insert.NUMBER, Insert.VARIABLE], # note
    "%I": [Insert.NUMBER, Insert.VARIABLE], # instrument
    
    "%h": [Insert.STRING_DROPDOWN], # sensor (value)
    "%H": [Insert.STRING_DROPDOWN], # sensor (bool)
    "%W": [Insert.STRING_DROPDOWN, Insert.VARIABLE], # motor direction # TODO
}
for accept_inserts in insert_kind.values():
    if Insert.STRING_DROPDOWN in accept_inserts:
        accept_inserts.append(Insert.STRING)
for accept_inserts in insert_kind.values():
    if Insert.VARIABLE in accept_inserts:
        accept_inserts.append(Insert.REPORTER)



def parse_color(hexcode):
    if hexcode:
        try:
            rgb = htmlcolor_parser.parse(hexcode)
            return Color.from_8bit(*rgb)
        except ValueError:
            pass
    return hexcode


def block_from_parts(parts, flag=None):
    # flag is currently unused
    
    text = ""
    arguments = []
    for part in parts:
        if isinstance(part, Insert):
            arguments.append(part)
        else:
            text += part
    
    script = None
    block = None
    poss_types = find_block(text)
    
    if poss_types:
        if len(poss_types) == 1:
            type = poss_types[0]
        else:            
            # verify insert types against block.inserts
            for type in poss_types:
                for (insert, need_type) in zip(arguments, type.inserts):
                    if need_type == '%f':
                        if insert.value not in math_functions:
                            break # Not this type.
                    else:
                        need_kinds = insert_kind[need_type]
                        if insert.kind not in need_kinds:
                            if insert.kind in (Insert.STRING, 
                                               Insert.STRING_DROPDOWN):
                                if Insert.NUMBER in need_kinds:
                                    # cast DROPDOWN / STRING to NUMBER
                                    if insert.value.strip().isdigit():
                                        continue # int
                                    elif value.strip().replace('.', '').isdigit():
                                        continue # float
                            
                                elif Insert.COLOR in need_kinds:
                                    # COLORs are STRINGs at this point
                                    continue
                            
                            elif (insert.kind == Insert.NUMBER and 
                                  Insert.STRING in need_kinds):
                                # cast NUMBER to STRING
                                continue
                            
                            
                            
                            break # Not this type.
                else:
                    break # Type is ok!
            else: # No types matched
                type = poss_types[0]
                print "WARNING: wrong args for '%s' block" % type.command
                print ' '.join(map(repr, parts))

        if type.command == "changeVariable":
            arguments.insert(1, None)
            # This gets replaced by a symbol from the blocks arguments:
            # either <#setVar:to:> or <#changeVar:by:>
        
        has_args = not type.command == "MouseClickEventHatMorph"
        allow_vars = not type.command in ("EventHatMorph", "KeyEventHatMorph")
        
        block_args = list(type.defaults)
        if has_args:
            for i in range(len(arguments)):
                insert = arguments[i]
                if insert is None:
                    continue
                
                kind = insert.kind
                if kind == Insert.VARIABLE and allow_vars:
                    arg = Block("readVariable", insert.value)
                else:
                    arg = insert.value
                
                if i < len(type.inserts):
                    insert_type = type.inserts[i]
                    need_kinds = insert_kind[insert_type]
                    
                    if kind in (Insert.STRING, 
                                Insert.STRING_DROPDOWN):
                        if insert_type == "%m":
                            arg = SpriteRef(arg)
                        
                        elif insert_type in ("%f", "%g", "%k"):
                            arg = str(arg) # Will not accept unicode!
                        
                        elif Insert.COLOR in need_kinds:
                                arg = parse_color(arg)
                    
                        elif Insert.NUMBER in need_kinds:
                            # cast DROPDOWN / STRING to NUMBER 
                            if arg.strip().isdigit():
                                arg = int(arg)
                            elif arg.strip().replace('.', '').isdigit():
                                arg = float(arg)
                    
                    elif ( Insert.STRING in need_kinds and
                           insert.kind == Insert.NUMBER ):
                        # cast NUMBER to STRING
                        arg = unicode(arg)
                
                if i >= len(block_args):
                    block_args.append(arg)
                else:
                    block_args[i] = arg

        block = Block(type, *block_args)

    if not block:
        e = "No block type found for %r \n"%text + repr(arguments)
        raise BlockError(e)

    return block


def variable_named(variable_name):
    if variable_name.endswith(' v'):
        variable_name = variable_name[:-2]
    
    # make sure this is a reporter!
    
    if variable_name.lower() in special_variables:
        value = special_variables[variable_name].copy()
        return Insert(Insert.SPECIAL, value)
    else:
        return Insert(Insert.VARIABLE, variable_name)



# Scripts

def p_file(t):
    """file : script"""
    t[0] = Script(None, blocks = t[1])

def p_script_end_newlines(t):
    """script : script NEWLINE"""
    t[0] = t[1]

def p_script_begin_newlines(t):
    """script : NEWLINE script"""
    t[0] = t[2]

def p_script(t):
    """script : script NEWLINE block
              | script NEWLINE block COMMENT"""
    block = t[3]
    if len(t) > 4:
        block.add_comment(t[4])
    t[0] = t[1] + [ block ]

def p_script_one(t):
    """script : block
              | block COMMENT
    """
    block = t[1]
    if len(t) > 2:
        block.add_comment(t[2])
    t[0] = [ block ]


# C blocks

def p_c_stack(t):
    """block : c_block  c_mouth END
             | if_block c_mouth END
    """
    block = t[1]
    block.args += [ t[2] ]
    t[0] = block #Block(block.command, *args)

def p_if_else(t):
    """block : if_block c_mouth ELSE c_mouth END
             | if_block c_mouth ELSE COMMENT c_mouth END"""
    block = t[1]
    block.type = blocks_by_cmd['doIfElse'][0]
    block.args += [ t[2] ]
    if len(t) > 6:
        block.args += [ t[5] ]
        block.add_comment(t[4])
    else:
        block.args += [ t[4] ]
    t[0] = block #Block(block.type, *args)

def p_c_mouth(t):
    """c_mouth : NEWLINE script NEWLINE"""
    t[0] = t[2]

def p_c_mouth_empty(t):
    """c_mouth : NEWLINE"""
    t[0] = None

def p_c_block(t):
    """c_block : c_block_def
               | c_block_def COMMENT
    """
    block = t[1]
    if len(t) > 2:
        block.add_comment(t[2])
    t[0] = block
    
def p_c_block_def(t):
    """c_block_def : FOREVER
                   | FOREVER IF insert
                   | REPEAT insert
                   | REPEAT UNTIL insert
    """
    t[0] = block_from_parts(t[1:])

def p_if_block(t):
    """if_block : if_block_def
                | if_block_def COMMENT
    """
    block = t[1]
    if len(t) > 2:
        block.add_comment(t[2])
    t[0] = block

def p_if_block_def(t):
    """if_block_def : IF insert"""
    block = block_from_parts(t[1:])
    t[0] = block


# Block

def p_block(t):
    """block : parts"""
    t[0] = block_from_parts(t[1])


# Parts

def p_parts(t):
    """parts : parts part"""
    t[0] = t[1] + [ t[2] ]

def p_part(t):
    """parts : part"""
    t[0] = [ t[1] ]


# Part

def p_insert_part(t):
    """part : insert"""
    t[0] = t[1]
    
def p_symbol_part(t):
    """part : SYMBOL"""
    t[0] = t[1]


# Inserts

def p_bool_insert(t):
    """insert : LBOOL parts RBOOL"""
    block = block_from_parts(t[2], "b")
    t[0] = Insert(Insert.BOOL, block)

def p_empty_bool_insert(t):
    """insert : LBOOL RBOOL"""
    t[0] = Insert(Insert.BOOL, False)

def p_string_insert(t):
    """insert : STRING"""
    (value, is_dropdown) = t[1]
    if value.lower() in special_strings:
        value = special_strings[value].copy()
        t[0] = Insert(Insert.SPECIAL, value)
    else:
        if is_dropdown:
            t[0] = Insert(Insert.STRING_DROPDOWN, value)
        else:
            t[0] = Insert(Insert.STRING, value)

def p_reporter_insert(t):
    """insert : LPAREN parts RPAREN"""
    parts = t[2]
    try: # Reporter block.
        block = block_from_parts(parts, "r")
        t[0] = Insert(Insert.REPORTER, block)
    
    except BlockError, e:
        # Maybe it's a variable? Check there are no args.
        for part in parts:
            if isinstance(part, Insert):
                raise e # Give up: not a block or a variable.
        
        # Variable!
        t[0] = variable_named(' '.join(parts))
        # TODO: allow multiple spaces in var names

def p_empty_reporter_insert(t):
    """insert : LPAREN RPAREN"""
    t[0] = Insert(Insert.REPORTER, None)
    
def p_number_insert(t):
    """insert : LPAREN number RPAREN
              | LPAREN number DROPDOWN
    """
    t[0] = Insert(Insert.NUMBER, t[2])
    
def p_variable_insert(t):
    """insert : LPAREN SYMBOL RPAREN
              | LPAREN SYMBOL DROPDOWN
    """
    variable_name = t[2]
    try: # Might be a reporter like 'costume #'
        block = block_from_parts(t[2], "r")
        t[0] = Insert(Insert.REPORTER, block)
    
    except BlockError:
        t[0] = variable_named(variable_name)
    
def p_number_int(t):
    """number : INT"""
    t[0] = t[1]

def p_number_float(t):
    """number : FLOAT"""
    t[0] = t[1]



# Override to make lt/gt < > work.
#
# This assumes )>( and )>( are never valid in any other context -- you can never
# have a boolean insert directly followed by a reporter, or vice versa.

def p_boolean_expr(t):
    """parts : insert LBOOL insert
             | insert RBOOL insert
    """
    t[0] = [ t[1], t[2], t[3] ]

    
# Keywords sometimes used in block text

def p_keyword_part(t):
    """part : IF
            | UNTIL
    """
    t[0] = t[1]



# Syntax errors.
def p_error(p):
    if p:
        err_msg = "Unexpected %(type)s: %(value)s " % p.__dict__
        pretty_error(p, err_msg)
    else:
        # EOF
        raise SyntaxError("invalid syntax")



# Build the parser.
path_to_file = os.path.join(os.getcwd(), __file__)
path_to_folder = os.path.split(path_to_file)[0]

block_plugin_parser = yacc.yacc(debug=DEBUG, write_tables=0) 
                               #outputdir=path_to_folder,

# TODO: figure out how to make pip/setup.py be nice to PLY
# so parsetab.py can be writeable




#coding=utf8

# Copyright © 2012 Tim Radvan
#
# This file is part of Kurt.
#
# Kurt is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Compiles a folder structure generated by decompile.py to a Scratch project.
Scripts are .txt files containing block plugin (scratchblocks) syntax.

    Usage: compile.py "path/to/project files/" """

import time
import os, sys

import codecs
def open(file, mode="r"):
    return codecs.open(file, mode, "utf-8")


try:
    import kurt
except ImportError: # try and find kurt directory
    path_to_file = os.path.join(os.getcwd(), __file__)
    path_to_lib = os.path.split(os.path.split(path_to_file)[0])[0]
    sys.path.append(path_to_lib)

from kurt import *



IGNORED_NAMES = [".DS_Store"]


class InvalidFile(Exception):
    def __init__(self, path, error):
        self.path = path
        self.error = error

        message = ("%s\n" % path) + unicode(error)
        Exception.__init__(self, message)


class FolderNotFoundException(Exception):
    pass

class FileExistsException(Exception):
    pass

class ParseError(Exception):
    pass



last_had_newline = True
def log(msg, newline=True):
    global last_had_newline
    if newline:
        if not last_had_newline:
            print
        print msg
    else:
        print msg,
    last_had_newline = newline


def split_filename_number(filename, path=None):
    parts = filename.split(' ')
    try:
        number = int(parts[0])
    except ValueError:
        if path:
            path = os.path.join(path, filename)
        else:
            path = filename
        raise InvalidFile(path, "Name must start with number")

    name = " ".join(parts[1:])

    return (number, name)



def read_costume_file(path):
    try:
        f = open(path)
    except IOError:
        return []

    line_count = 0
    costumes = []
    while 1:
        line = f.readline()
        line_count += 1
        if not line:
            return costumes

        line = line.strip()
        if line:
            filename = line.strip()
            costume = {
                "filename": filename,
                "rotationCenter": None,
            }

            while 1:
                line = f.readline()

                if not line.startswith("#"):
                    line_count += 1
                    if not line:
                        return costumes

                    line = line.strip()
                    if not line:
                        break

                    parts = line.split(":")
                    if len(parts) == 2:
                        (key, value) = parts
                        costume[key] = value
                    elif len(parts) == 1:
                        key = parts[0]
                        costume[key] = None
                    else:
                        raise InvalidFile(path, "Line %i" % line_count)

            costumes.append(costume)




def read_script_file(morph, path):
    file = open(path)

    settings = {
        "pos": "(20, 20)",
    }

    while 1:
        try:
            line = file.readline()
        except EOFError:
            return
        line = line.strip()

        if line:
            try:
                (name, value) = line.split(":")
            except ValueError:
                raise InvalidFile(path, "invalid line: "+line)

            name = name.strip().lower()
            value = value.strip()
            for setting in settings.keys():
                if name.startswith(setting):
                    settings[setting] = value
                    break
            else:
                settings[name] = value

        else:
            break

    pos = Point.from_string(settings["pos"])

    data = ""
    while 1:
        more_data = file.read()
        if not more_data: break
        data += more_data

    data = unicode(data)
    data = data.replace('\r\n', '\n')
    data = data.replace('\r', '\n')

    script = parse_block_plugin(data)
    script.pos = pos
    script.morph = morph
    # TODO: move position info inside parser?

    file.close()

    return script



def import_sprite(project_dir, sprite_name):
    sprite_dir = os.path.join(project_dir, sprite_name)
    is_stage = (sprite_name in ("Stage", "00 Stage"))
    if is_stage:
        sprite_number = 0
    else:
        try:
            (sprite_number, sprite_name) \
                = split_filename_number(sprite_name, project_dir)
        except InvalidFile:
            sprite_number = None
            sprite_name = sprite_name

    log("* "+sprite_name, False)
    start_time = time.time()

    if is_stage:
        sprite = Stage()
        stage_background = sprite.backgrounds[0]
        sprite.backgrounds = []
    else:
        sprite = Sprite()
        sprite.name = sprite_name


    # Variables
    var_list_path = os.path.join(sprite_dir, "variables.txt")
    if os.path.exists(var_list_path):
        var_file = open(var_list_path)
        for line in var_file:
            line = line.strip("\r\n")
            if line:
                parts = line.split(" = ")
                var_name = parts[0]
                
                if find_block(var_name):
                    log("WARNING: invalid variable name: %s" % var_name)
                
                value = " = ".join(parts[1:])
                sprite.variables[var_name] = value


    # Lists
    lists_dir = os.path.join(sprite_dir, "lists")
    if os.path.exists(lists_dir):
        list_names = os.listdir(lists_dir)
        for list_name in list_names:
            if list_name in IGNORED_NAMES:
                continue
            list_path = os.path.join(lists_dir, list_name)

            if "." in list_name: # strip extension
                list_name = ".".join(list_name.split(".")[:-1])

            list_file = open(list_path)
            items = [line.strip("\r\n") for line in list_file.readlines()]
            sprite.lists[list_name] \
                = ScratchListMorph(name=list_name, items=items)
            list_file.close()


    # Scripts
    scripts_dir = os.path.join(sprite_dir, "scripts")
    if os.path.exists(scripts_dir):
        script_names = os.listdir(scripts_dir)

        for script_name in script_names:
            script_path = os.path.join(scripts_dir, script_name)
            script = read_script_file(sprite, script_path)
            if script:
                sprite.scripts.append(script)

        sprite.scripts.sort(key=lambda script: script.pos.y)

    # Costumes/Backgrounds
    if is_stage:
        costumes_dir = os.path.join(sprite_dir, "backgrounds")
    else:
        costumes_dir = os.path.join(sprite_dir, "costumes")

    costumes = []
    selected_costume = None
    if os.path.exists(costumes_dir):
        costume_file = costumes_dir+".txt"
        if os.path.exists(costume_file):
            costumes = read_costume_file(costume_file)

        found_costumes = os.listdir(costumes_dir)
        for ignore in IGNORED_NAMES:
            while ignore in found_costumes:
                found_costumes.remove(ignore)
        found_costumes.sort()

        remove_costumes = []
        for costume in costumes:
            costume_path = os.path.join(costumes_dir, costume["filename"])
            if ( costume["filename"] not in found_costumes or
                 not os.path.exists(costume_path) ):
                log("  - " + "Couldn't find costume: "+costume["filename"])
                remove_costumes.append(costume)
        for costume in remove_costumes:
            costumes.remove(costume)


        for filename in found_costumes:
            for other in costumes:
                if filename == other["filename"]:
                    break
            else:
                costumes.append({
                    "filename": filename,
                    "rotationCenter": None,
                })

        for costume in costumes:
            try:
                (number, name) = split_filename_number(costume["filename"])
                costume["number"] = number
            except InvalidFile:
                costume["number"] = None

        costumes.sort(key=lambda c: c["number"])
        costumes.sort(key=lambda c: c["number"] is None) # sort new costumes to end

        for costume_args in costumes:
            filename = costume_args["filename"]
            log("  - " + filename)

            costume_path = os.path.join(costumes_dir, filename)
            costume = Image.load(costume_path)
            if not costume:
                raise InvalidFile(costume_path, "Couldn't load image")

            if "name" in costume_args and costume_args["name"]:
                costume.name = costume_args["name"]
            else:
                try:
                    (_, costume.name) = split_filename_number(costume.name)
                except InvalidFile:
                    pass

            if costume_args["rotationCenter"]:
                costume.rotationCenter = Point.from_string(
                    costume_args["rotationCenter"])
            else:
                try:
                    size = costume.size
                except ValueError:
                    costume.rotationCenter = Point(0, 0)
                    size = None
                if size:
                    (width, height) = size
                    costume.rotationCenter = Point(int(width / 2), int(height / 2))

            if "selected" in costume_args:
                selected_costume = costume

            sprite.images.append(costume)

    if is_stage and not costumes:
        sprite.backgrounds = [stage_background]
        selected_costume = stage_background

    if not selected_costume and sprite.images:
        selected_costume = sprite.images[0]

    sprite.costume = selected_costume


    # Finished!
    sprite_save_time = time.time() - start_time
    log(sprite_save_time)

    return (sprite_number, sprite)




def compile(project_dir, debug=True): # DEBUG: set to false
    start_time = time.time()

    if project_dir.endswith(".sb"):
        project_dir = project_dir[:-3]

    if project_dir.endswith(" files"):
        project_path = project_dir[:-6]

    else:
        project_path = project_dir
        project_dir += " files"

    if not os.path.exists(project_dir):
        raise FolderNotFoundException(project_dir)

    project = ScratchProjectFile.new(project_path)

    if os.path.exists(project.path):
        raise FileExistsException(project.path)
    
    # Sprites
    log("Importing sprites...")

    sprite_names = []
    for name in os.listdir(project_dir):
        if os.path.isdir(os.path.join(project_dir, name)):
            sprite_names.append(name)
    
    if "00 Stage" in sprite_names:
        stage_name = "00 Stage"
    elif "Stage" in sprite_names:
        stage_name = "Stage"
    else:
        stage_name = None

    if stage_name:
        sprite_names.remove(stage_name)

    for ignore in IGNORED_NAMES:
        while ignore in sprite_names:
            sprite_names.remove(ignore)

    if stage_name is not None:
        (_, stage) = import_sprite(project_dir, stage_name)
        project.stage = stage

    sprites = [import_sprite(project_dir, name) for name in sprite_names]
    sprites.sort(key=lambda (n, s): n)
    sprites.sort(key=lambda (n, s): n is None) # sort new sprites to end
    for (number, sprite) in sprites:
        project.sprites.append(sprite)
    
    # Check variables
    undefined_vars = set()
    for sprite in project.sprites:
        for script in sprite.scripts:
            script.replace_sprite_refs(lookup_sprite_named = project.get_sprite)

            for var in script.find_undefined_variables(project.stage):
                undefined_vars.add(var)

    log("")
    for var in undefined_vars:
         log("WARNING: variable not found: %s" % var)
    
    # Thumbnail
    thumb_path = os.path.join(project_dir, "thumbnail.png")
    if os.path.exists(thumb_path):
        log("")
        log("Importing thumbnail...")
        project.info["thumbnail"] = Image.load(thumb_path)
    
    # Notes
    notes_path = os.path.join(project_dir, "notes.txt")
    if os.path.exists(notes_path):
        notes = open(notes_path).read()
        notes = notes.replace("\r\n", "\n")
        if "kurt" not in notes.lower():
            notes += "\n\n---\n\n"
            notes += ScratchProjectFile.DEFAULT_COMMENT
        project.info["comment"] = notes
    
    # Done!
    compile_time = time.time() - start_time
    log("Compiled! %f" % compile_time)

    return project



def cmd_compile(path):
    """Usage: compile.py "path/to/project files/" 
    
    folder structure -> scratch project"""
    
    if path.endswith(".sb"):
        path = path[:-3]
    if path.endswith(" files"):
        path = path[:-6]

    project = None
    try:
        compile_time = time.time()

        project = compile(path)
        log("")

        log("Saving...")
        save_time = time.time()
        project.save()
        log("Saved! %f" % (time.time() - save_time))

        log("")
        log("Total %f secs" % (time.time() - compile_time))

    except FileExistsException, e:
        log("File exists: %s" % unicode(e))

    except InvalidFile, e:
        log("")
        log("Invalid file: %s" % e)

    except FolderNotFoundException, e:
        log("Folder missing: %s" % unicode(e))

    return project # useful for debugging



if __name__ == '__main__':
    if len(sys.argv) <= 1:
        print __doc__
        exit()

    else:
        cmd_compile(sys.argv[1])





#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""ObjTable - for de/serialising an object and its related objects.

(You probably want to use the classes in kurt.files directly.)

Otherwise, the main class in this file is ObjTable.
"""

from construct import *
from functools import partial
import inspect

from inline_objects import Field, Ref
from fixed_objects import *
import fixed_objects
from user_objects import *
import user_objects


class ObjectAdapter(Adapter):
    """Decodes a construct to a pythonic class representation.
    The class must have a from_construct classmethod and a to_construct 
    instancemethod.
    """
    def __init__(self, classes, *args, **kwargs):
        """Initialize an adapter for a new type/object(s).
        @param classes: class, list of classes, or dict of obj.classID name to 
        class mapping.
            eg ObjectAdapter({"String": String, "Array": Collection}, <subcon>)
        Note: Must use new-style objects, ie. subclasses of object.
        """
        Adapter.__init__(self, *args, **kwargs)
        
        if isinstance(classes, list):
            classes = dict((cls.__name__, cls) for cls in classes)
        self.classes = classes
    
    def _get_class(self, classID):
        if inspect.isclass(self.classes):
            return self.classes
        else:
            return self.classes[classID]
    
    def _encode(self, obj, context):
        """Encodes a class to a lower-level object using the class' own 
        to_construct function.
        If no such function is defined, returns the object unchanged.
        """
        func = getattr(obj, 'to_construct', None)
        if callable(func):
            return func(context)
        else:
            return obj
    
    def _decode(self, obj, context):
        """Initialises a new Python class from a construct using the mapping 
        passed to the adapter.
        """
        cls = self._get_class(obj.classID)
        return cls.from_construct(obj, context)


def obj_classes_from_module(module):
    """Return a list of classes in a module that have a 'classID' attribute."""
    for name in dir(module):
        if not name.startswith('_'):
            cls = getattr(module, name)
            if getattr(cls, 'classID', None):
                yield (name, cls)



### Fixed-format objects ###

fixed_object_classes = []
fixed_object_ids_by_name = {}
fixed_object_cons_by_name = {}

for (name, cls) in obj_classes_from_module(fixed_objects):
    fixed_object_classes.append(cls)
    fixed_object_ids_by_name[name] = cls.classID
    fixed_object_cons_by_name[name] = cls._construct

FixedObjectAdapter = partial(ObjectAdapter, fixed_object_classes)       

fixed_object = FixedObjectAdapter(Struct("fixed_object",
    Enum(UBInt8("classID"), **fixed_object_ids_by_name),
    Switch("value", lambda ctx: ctx.classID, fixed_object_cons_by_name),
))
fixed_object.__doc__ = """Construct for FixedObjects.
Stored in the object table. May contain references."""


### User-class objects ###

user_object_classes = []
user_object_ids_by_name = {}

for (name, cls) in obj_classes_from_module(user_objects):
    user_object_classes.append(cls)
    user_object_ids_by_name[name] = cls.classID

UserObjectAdapter = partial(ObjectAdapter, user_object_classes)

user_object = UserObjectAdapter(Struct("user_object",
    Enum(UBInt8("classID"),
        **user_object_ids_by_name
    ),
    UBInt8("version"),
    UBInt8("length"),
    Rename("field_values", MetaRepeater(lambda ctx: ctx.length, Field)),
))
user_object.__doc__ = """Construct for UserObjects.
Stored in the object table. May contain references."""


### Object Table ###

class PythonicAdapter(Adapter):
    """Converts from FixedObject classes to native Python types.
    * String -- python str
    * UTF8 -- python unicode
    * Dictionary -- dict
    * Array -- list/tuple
    """
    def _encode(self, obj, context):
        if isinstance(obj, str):
            return String(obj)
        elif isinstance(obj, unicode):
            return UTF8(obj)
        elif isinstance(obj, dict):
            return Dictionary(obj)
        elif isinstance(obj, list) or isinstance(obj, tuple):
            return Array(obj)
        elif isinstance(obj, Script):
            return Array(obj.to_array())
        else:
            return obj
    
    def _decode(self, obj, context):
        if isinstance(obj, String):
            return str(obj.value)
        elif isinstance(obj, UTF8):
            return unicode(obj.value)
        elif isinstance(obj, Dictionary):
            return obj.value
        elif isinstance(obj, Array):
            return obj.value
        else:
            return obj

class ObjectAdapter(Adapter):
    def _encode(self, obj, context):
        classID = obj.classID
        if classID in fixed_object_ids_by_name:
            classID = fixed_object_ids_by_name[classID]
        elif classID in user_object_ids_by_name:
            classID = user_object_ids_by_name[classID]
        return Container(
            classID = classID,
            object = obj,
        )
    
    def _decode(self, obj, context):
        return obj.object

_obj_table_entry = PythonicAdapter(ObjectAdapter(Struct("object",
    Peek(UBInt8("classID")),
    IfThenElse("object", lambda ctx: ctx.classID < 99,
        fixed_object,
        user_object,
    ),
)))
_obj_table_entry.__doc__ = """Construct for object table entries, both 
    UserObjects and FixedObjects."""


class ObjectTableAdapter(Adapter):
    def _encode(self, objects, context):
        return Container(
            header = "ObjS\x01Stch\x01",
            length = len(objects),
            objects = objects,
        )
    
    def _decode(self, table, context):
        assert table.length == len(table.objects), "File corrupt?"
        return table.objects


class ObjectNetworkAdapter(Adapter):
    """Object network <--> object table listing objects containing Refs"""
    def _encode(self, root, context):
        orig_objects = []
        
        def get_ref(value):            
            """Returns the index of the given object in the object table, 
            adding it if needed.
            """
            objects = self._objects
            
            value = PythonicAdapter(Pass)._encode(value, context)
            # Convert strs to FixedObjects here to make sure they get encoded 
            # correctly
            
            if isinstance(value, UserObject) or isinstance(value, FixedObject):
                if getattr(value, '_tmp_index', None):
                    index = value._tmp_index
                else:
                    objects.append(value)
                    index = len(objects)
                    value._tmp_index = index
                    orig_objects.append(value) # save the object so we can
                                               # strip the _tmp_indexes later
                return Ref(index)
            else:
                return value # Inline value
        
        def fix_fields(obj):
            obj = PythonicAdapter(Pass)._encode(obj, context)
            # Convert strs to FixedObjects here to make sure they get encoded
            # correctly
            
            if isinstance(obj, UserObject):
                fixed_obj = obj.to_construct(Container())
                fixed_obj.field_values = [get_ref(value) 
                                          for value in fixed_obj.field_values]
                #fixed_obj = obj.__class__(field_values, version = obj.version)
            
            elif isinstance(obj, Dictionary):
                fixed_obj = obj.__class__(dict(
                    (get_ref(field), get_ref(value))
                    for (field, value) in obj.value.items()
                ))
                
            elif isinstance(obj, dict):
                fixed_obj = dict(
                    (get_ref(field), get_ref(value)) 
                    for (field, value) in obj.items()
                )
                
            elif isinstance(obj, list):
                fixed_obj = [get_ref(field) for field in obj]
                
            elif isinstance(obj, Form):
                fixed_obj = obj.__class__(**dict(
                    (field, get_ref(value)) 
                    for (field, value) in obj.value.items()
                ))
                
            elif isinstance(obj, ContainsRefs):
                fixed_obj = obj.__class__([get_ref(field) 
                                           for field in obj.value])
                
            else:
                return obj
            
            fixed_obj._made_from = obj
            return fixed_obj
        
        root = PythonicAdapter(Pass)._encode(root, context)
        
        i = 0
        self._objects = objects = [root]
        root._tmp_index = 1
        while i < len(objects):
            objects[i] = fix_fields(objects[i])
            i += 1
        
        for obj in orig_objects:
            obj._tmp_index = None
            # Strip indexes off objects in case we want to save again later
        
        return objects
    
    def _decode(self, objects, context):        
        def resolve_ref(obj, objects=objects):
            if isinstance(obj, Ref):
                # first entry is 1
                return objects[obj.index - 1]
            else:
                return obj
        
        # Reading the ObjTable backwards somehow makes more sense.
        for i in xrange(len(objects)-1, -1, -1):
            obj = objects[i]
            
            if isinstance(obj, UserObject):
                for field_name in obj.fields:
                    value = obj.fields[field_name]
                    value = resolve_ref(value)
                    obj.fields[field_name] = value
            
            elif isinstance(obj, Dictionary):
                obj.value = dict(
                    (resolve_ref(field), resolve_ref(value)) 
                    for (field, value) in obj.value.items()
                )
            
            elif isinstance(obj, dict):
                obj = dict(
                    (resolve_ref(field), resolve_ref(value)) 
                    for (field, value) in obj.items()
                )
            
            elif isinstance(obj, list):
                obj = [resolve_ref(field) for field in obj]
            
            elif isinstance(obj, Form):
                for field in obj.value:
                    value = getattr(obj, field)
                    value = resolve_ref(value)
                    setattr(obj, field, value)
            
            elif isinstance(obj, ContainsRefs):
                obj.value = [resolve_ref(field) for field in obj.value]
            
            objects[i] = obj
        
        for obj in objects:
            if isinstance(obj, UserObject) or isinstance(obj, Form):
                obj.built()
        
        root = objects[0]
        return root

_obj_table_entries = ObjectTableAdapter(Struct("object_table",
    Const(Bytes("header", 10), "ObjS\x01Stch\x01"),
    UBInt32("length"),
    Rename("objects", MetaRepeater(lambda ctx: ctx.length, _obj_table_entry)),
))

ObjTable = ObjectNetworkAdapter(_obj_table_entries)
ObjTable.__doc__ = """Construct for parsing a binary object table to pythonic 
    object(s).
    """ #Includes "ObjS\\x01Stch\\x01" header.

class InfoTableAdapter(Subconstruct):
    """Info ObjTable found in the project header.
    Adds the preceding info_size header (4 bytes).
    
    Parses to a Dictionary. Includes the following keys:
        thumbnail - image showing a small picture of the stage when the project 
                    was saved
        author - name of the user who saved or shared this project
        comment - author's comments about the project
        history - a string containing the project save/upload history
        scratch-version - the version of Scratch that saved the project
    """
    info_size = UBInt32("info_size")
    
    def _parse(self, stream, context):
        self.info_size._parse(stream, Container())
        objtable = self.subcon._parse(stream, context)
        return objtable
    
    def _build(self, obj, stream, context):
        bytes = self.subcon.build(obj)
        size = len(bytes)
        stream.write(self.info_size.build(size))
        stream.write(bytes)
        
InfoTable = InfoTableAdapter(ObjTable)



ScratchStageMorph = Stage
ScratchSpriteMorph = Sprite
ImageMedia = Image
SoundMedia = Sound

__all__ = [
    'ObjTable', '_obj_table_entry', '_obj_table_entries', 
    'UserObject', 'FixedObject', 'Field', 
    'ScratchStageMorph', 'ScratchSpriteMorph',
] + [cls.__name__ for cls in fixed_object_classes + user_object_classes]



from kurt import *
from kurt.scratchblocks.parser import math_functions

p = parse_scratchblocks

assert p("say (length of [hello])")[0].args[0].command == 'stringLength:' 
assert p("say (length of (var))")[0].args[0].command == 'stringLength:' 
assert p("say (length of [list v])")[0].args[0].command == 'lineCountOfList:'

assert p(u'[x position v] of [Sprite1]')[0].command == 'getAttribute:of:'
assert p(u'[x position v] of [Sprite1 v]')[0].command == 'getAttribute:of:'
assert p(u'[sqrt v] of (6)')[0].command == 'computeFunction:of:'
assert p(u'[sqrt v] of [6]')[0].command == 'computeFunction:of:'

assert p(u'[x position v] of (var)')[0].command == 'getAttribute:of:'
assert p(u'[sqrt v] of (var)')[0].command == 'computeFunction:of:'
for mf in math_functions:
    assert p(u'[%s v] of (var)' % mf)[0].command == 'computeFunction:of:'

assert p('set pen color to [#faa41a]')[0] == Block('penColor:', Color(1000, 656, 104))
assert p('set pen color to (0)')[0] == Block('setPenHueTo:', 0)

assert p("when I receive (cheese)")[0] == Block('EventHatMorph', 'cheese')


#coding=utf8

# Copyright © 2012 Tim Radvan
#
# This file is part of Kurt.
#
# Kurt is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""User-class objects with variable numbers of fields.
Most of the objects you're interested in live here.

They support dot notation for accessing fields. Use .fields.keys() to see
available fields [dir() won't show them.]
"""

from construct import Container
import os
import StringIO
from array import array

try:
    import PIL.Image
except ImportError:
    PIL = None

from inline_objects import Ref
from fixed_objects import *


def require_pil():
    if not PIL:
        raise ValueError, "Missing dependency: " \
            "PIL library needed for image support"



class UserObject(object):
    """A user-class object with a variable number of fields.
    Supports dot notation for accessing fields.
    Use .fields.keys() to see available fields [dir() won't show them.]

    Each class lists its field order in _fields.
    Unknown fields not in this list are named "undefined-%i", where i is the
    field index.
    """
    _fields = []
    _version = 1

    def to_construct(self, context):
        field_values = self.field_values[:]

        for i in range(len(field_values)):
            value = field_values[i]
            if i < len(self._fields):
                field = self._fields[i]
                field_values[i] = self._encode_field(field, value)

        return Container(
            classID = self.__class__.__name__,
            field_values = field_values,
            length = len(field_values),
            version = self.version,
        )

    def _encode_field(self, name, value):
        """Modify the field with the given name before saving, if necessary.
        Override this in subclass to modify building of specific fields.
        """
        return value

    @classmethod
    def from_construct(cls, obj, context):
        return cls(obj.field_values, version=obj.version)

    def _decode_field(cls, name, value):
        """Return value of named field passed to object's constructor.
        Override this in subclass to modify specific fields.
        """
        return value

    def set_defaults(self):
        """Set defaults on self. Return nothing.
        Subclasses can override this to setup default values.
        """
        pass

    def built(self):
        for field in self._fields:
            if field in self.fields:
                value = self.fields[field]
                self.fields[field] = self._decode_field(field, value)

    def __init__(self, field_values=None, **args):
        """Initalize a UserObject.
        @param field_values: (optional) list of fields as parsed from a file.
        @param **args: field values.
        """
        self.fields = dict(zip(self._fields, [None] * len(self._fields)))

        self.version = self._version
        if 'version' in args:
            self.version = args.pop('version')

        self.set_defaults()

        if field_values:
            defined_fields = self._fields[:]
            defined_fields += tuple("undefined-%i" % i
                for i in range(len(defined_fields), len(field_values)))
            self.fields.update(zip(defined_fields, field_values))

        self.fields.update(args)

        if "name" in self.fields:
            # make sure we use property setter
            name = self.fields["name"]
            del self.fields["name"]
            self.name = name

    def __getattr__(self, name):
        if name in self.fields:
            return self.fields[name]
        else:
            raise AttributeError, ('%s instance has no attribute %s'
                % (self.__class__.__name__, name))

    def __setattr__(self, name, value):
        if name in self._fields:
            self.fields[name] = value
        else:
            object.__setattr__(self, name, value)

    def __delattr__(self, name):
        if name in self.fields:
            del self.fields[name]
        else:
            object.__delattr__(self, name)

    def __len__(self):
        return len(self.fields)

    @property
    def ordered_fields(self):
        ordered_fields = []
        fields = self.fields.copy()
        for field_name in self._fields:
            value = None
            if field_name in self.fields:
                value = fields.pop(field_name)
            ordered_fields.append((field_name, value))

        # leftover undefined fields
        fields = sorted(fields.items(),
            key=lambda (field,value): int(field.split('-')[1])
        )
        for field_name, value in fields:
            ordered_fields.append((field_name, value))

        return ordered_fields

    @property
    def field_values(self):
        return [value for (field_name, value) in self.ordered_fields]

    def __repr__(self):
        name = getattr(self, "name", "")
        return "<%s(%s)>" % (self.__class__.__name__, name)



### Squeak & Morphic classes ###
class BaseMorph(UserObject):
    _fields = ("bounds", "owner", "submorphs", "color", "flags", "properties")

    def set_defaults(self):
        self.flags = 0
        self.submorphs = []
        self.color = Color(1023, 1023, 1023)

class Morph(BaseMorph):
    """Base class for most UserObjects."""
    classID = 100

class BorderedMorph(BaseMorph):
    classID = 101
    _fields = Morph._fields + ("borderWidth", "borderColor")

    def set_defaults(self):
        BaseMorph.set_defaults(self)
        self.borderWidth = 1
        self.borderColor = Color(0, 0, 0)

class RectangleMorph(BorderedMorph):
    classID = 102

class EllipseMorph(BorderedMorph):
    classID = 103

class AlignmentMorph(RectangleMorph):
    classID = 104
    _fields = RectangleMorph._fields + ("orientation", "centering", "hResizing",
        "vResizing", "inset")

class StringMorph(BaseMorph):
    classID = 105
    _fields = Morph._fields + ("font_with_size", "emphasis", "contents")

class UpdatingStringMorph(StringMorph):
    classID = 106

class SimpleSliderMorph(BorderedMorph):
    classID = 107
    _fields = BorderedMorph._fields + ("slider", "value", "setValueSelector",
        "sliderShadow", "sliderColor", "descending", "model", "target",
        "actionSelector", "arguments", "actWhen")

class SimpleButtonMorph(RectangleMorph):
    classID = 108
    _fields = RectangleMorph._fields + ("target", "actionSelector", "arguments",
        "actWhen")

class SampledSound(UserObject):
    classID = 109
    _fields = ("envelopes", "scaledVol", "initialCount", "samples",
        "originalSamplingRate", "samplesSize", "scaledIncrement",
        "scaledInitialIndex")

class ImageMorph(BaseMorph):
    classID = 110
    _fields = Morph._fields + ("form", "transparency")

class SketchMorph(BaseMorph):
    classID = 111
    _fields = Morph._fields + ("originalForm", "rotationCenter",
        "rotationDegrees", "rotationStyle", "scalePoint", "offsetWhenRotated")




### Scratch-specific classes ###

class ScriptableScratchMorph(BaseMorph):
    _fields = Morph._fields + ("name", "variables", "scripts", "isClone", "media",
        "costume")

    def __init__(self, *args, **kwargs):
        UserObject.__init__(self, *args, **kwargs)

        self.images = []
        self.sounds = []

        self.build_media() # returns silently if self.media is still a Ref

    def set_defaults(self):
        BaseMorph.set_defaults(self)

        self.scripts = ScriptCollection()
        self.media = []
        self.costume = None # defaults to first Image in self.media on save
        self.variables = {}
        self.lists = {}
        self.isClone = False

        self.volume = 100
        self.tempoBPM = 60

    def built(self):
        UserObject.built(self)

        scripts = [Script.from_array(self, script) for script in self.scripts]
        self.scripts = ScriptCollection(scripts)

        comments = []
        for script in self.scripts:
            if isinstance(script, Comment):
                comments.append(script)
        for comment in comments:
            self.scripts.remove(comment)

        blocks_by_id = list(self.blocks_by_id())

        for comment in comments:
            comment.attach_scripts(blocks_by_id)

        self.build_media()

    def build_media(self):
        if isinstance(self.media, Ref):
            return # Don't run this yet!

        media = self.media
        self.media = []
        for media in media:
            if isinstance(media, Sound):
                self.sounds.append(media)
            elif isinstance(media, Image):
                self.images.append(media)
            else:
                self.media.append(media)

    def blocks_by_id(self):
        """Return a list of all the blocks in script order but reverse script
        blocks order.
        Used to determine which block a Comment is anchored to.

        Note that Squeak arrays are 1-based, so index with:
            blocks_by_id[index - 1]
        """
        for script in self.scripts:
            for block in reversed(list(script.to_block_list())):
                yield block

    def normalize(self):
        """Called before saving"""
        if not self.costume:
            for media in self.media + self.images:
                if isinstance(media, Image):
                    self.costume = media
                    break
            else:
                raise ValueError("%r does not have a costume" % self)

        self.lists = dict(
            (unicode(name), list) for (name, list) in self.lists.items()
        )
        for list_name in self.lists:
            scratch_list = self.lists[list_name]
            if not isinstance(scratch_list, ScratchListMorph):
                scratch_list = ScratchListMorph(items=scratch_list)
                self.lists[list_name] = scratch_list
            scratch_list.name = list_name

            # This would show the list watcher on the stage:
            #scratch_list.target = self
            #if isinstance(self, Stage):
            #     scratch_list.owner = self
            #else:
            #    scratch_list.owner = self.owner

            scratch_list.normalize()

    def _encode_field(self, name, value):
        if name == 'scripts':
            scripts = [script.to_array() for script in value]
            blocks_by_id = list(self.blocks_by_id())
            for block in blocks_by_id:
                if block.comment:
                    scripts.append(block.comment.to_array(blocks_by_id))
            return scripts
        elif name == 'media':
            return OrderedCollection(self.sounds + self.images + self.media)
        else:
            return value


class SensorBoardMorph(BaseMorph):
    classID = 123
    _fields = BaseMorph._fields + ("unknown",)
                                  # TODO — I have NO idea what this does.


class Sprite(ScriptableScratchMorph):
    """A sprite.
    Main attributes:
        scripts
        variables
        lists
        costumes
        sounds
    Use .fields.keys() to see all available fields.
    """
    classID = 124
    _fields = ScriptableScratchMorph._fields + ("visibility", "scalePoint",
        "rotationDegrees", "rotationStyle", "volume", "tempoBPM", "draggable",
        "sceneStates", "lists")
    _version = 3

    def set_defaults(self):
        ScriptableScratchMorph.set_defaults(self)

        self.name = "Sprite1"
        self.color = Color(0, 0, 1023)
        # self.owner — Stage
        # self.bounds = Rectangle() - default to size of costume?

        self.visibility = 100
        self.scalePoint = Point(1.0, 1.0)
        self.rotationDegrees = 0.0
        self.rotationStyle = Symbol("normal")
        self.draggable = False
        self.sceneStates = {}

    def normalize(self):
        """Called before saving"""
        ScriptableScratchMorph.normalize(self)

        if not self.bounds:
            try:
                self.bounds = Rectangle(
                    [0, 0, self.costume.width, self.costume.height])
            except ValueError:
                # invalid costume, or maybe JPG
                self.bounds = Rectangle([0, 0, 100, 100])

    @property
    def costumes(self):
        return self.images

    @costumes.setter
    def costumes(self, value):
        self.images = value


class SpriteCollection(OrderedCollection):
    def __getitem__(self, item):
        try:
            index = int(item)
        except ValueError:
            for sprite in self.value:
                if sprite.name == item:
                    return sprite
        return self.value[index]

    def __repr__(self):
        return repr(self.value)


class Stage(ScriptableScratchMorph):
    """The project stage. Contains project contents including sprites and media.
    Main attributes:
        sprites - ordered list of sprites.
        submorphs - everything on the stage, including sprites &
                    variable/list watchers.
        scripts
        variables
        lists
        backgrounds
        sounds
    Use .fields.keys() to see all available fields.
    """
    classID = 125
    _fields = ScriptableScratchMorph._fields + ("zoom", "hPan", "vPan",
        "obsoleteSavedState", "sprites", "volume", "tempoBPM", "sceneStates",
        "lists")
    _version = 5

    def set_defaults(self):
        ScriptableScratchMorph.set_defaults(self)

        self.name = "Stage"
        self.bounds = Rectangle([0, 0, 480, 360])
        self.color = Color(1023, 1023, 1023)

        self.zoom = 1.0
        self.hPan =  0
        self.vPan =  0
        self.sprites = SpriteCollection()
        self.sceneStates = {}

        image = Image(
            name = "background",
            form = ColorForm(
                width = 480,
                height = 360,
                depth = 1,
                bits = ByteArray("\xf5\x18\xff\x00\x00Ta\x00"),
            ),
        )

        self.media = [image]
        self.images = [image]
        self.costume = image

    def normalize(self):
        """Called before saving"""
        ScriptableScratchMorph.normalize(self)

        for sprite in self.sprites:
            if sprite not in self.submorphs:
                self.submorphs.append(sprite)
            sprite.owner = self
            sprite.normalize()

    def built(self):
        ScriptableScratchMorph.built(self)
        self.sprites = SpriteCollection(self.sprites)

    def _encode_field(self, name, value):
        value = ScriptableScratchMorph._encode_field(self, name, value)

        if name == 'sprites':
            return OrderedCollection(self.sprites)
        else:
            return value

    @property
    def background(self):
        return self.costume

    @background.setter
    def background(self, value):
        self.costume = value

    @property
    def backgrounds(self):
        return self.images

    @backgrounds.setter
    def backgrounds(self, value):
        self.images = value


from scripts import Script, Comment, ScriptCollection
# Yes, this is stupid. Circular dependencies ftw. -_-




class ChoiceArgMorph(BaseMorph):
    """unused?"""
    classID = 140

class ColorArgMorph(BaseMorph):
    """unused?"""
    classID = 141

class ExpressionArgMorph(BaseMorph):
    """unused?"""
    classID = 142

class SpriteArgMorph(BaseMorph):
    """unused?"""
    classID = 145

class BlockMorph(BaseMorph):
    """unused?"""
    classID = 147
    _fields = Morph._fields + ("isSpecialForm", "oldColor")

class CommandBlockMorph(BlockMorph):
    """unused?"""
    classID = 148
    _fields = BlockMorph._fields + ("commandSpec", "argMorphs", "titleMorph",
        "receiver", "selector", "isReporter", "isTimed", "wantsName",
        "wantsPossession")

class CBlockMorph(BaseMorph):
    """unused?"""
    classID = 149

class HatBlockMorph(BaseMorph):
    """unused?"""
    classID = 151

class ScratchScriptsMorph(BorderedMorph):
    classID = 153

    def __iter__(self):
        return iter(self.submorphs)

class ScratchSliderMorph(BaseMorph):
    """unused?"""
    classID = 154

class WatcherMorph(AlignmentMorph):
    """A variable watcher."""
    classID = 155
    _fields = AlignmentMorph._fields + ("titleMorph", "readout", "readoutFrame",
        "scratchSlider", "watcher", "isSpriteSpecific", "unused", "sliderMin",
        "sliderMax", "isLarge")
    _version = 5

    @property
    def name(self):
        return self.titleMorph.contents

    @name.setter
    def name(self, value):
        self.titleMorph.contents = value

class SetterBlockMorph(BaseMorph):
    """unused?"""
    classID = 157

class EventHatMorph(BaseMorph):
    """unused?"""
    classID = 158

class VariableBlockMorph(CommandBlockMorph):
    """unused?"""
    classID = 160
    _fields = CommandBlockMorph._fields + ("isBoolean",)




class ScratchMedia(UserObject):
    _fields = ("name",)


class Image(ScratchMedia):
    """An image file, used for costumes and backgrounds.

    You can't modify image data in-place (excepting `textBox`) -- create a new
    image object using load() or from_image() instead.

    Class methods:
        load(path) — load a PNG or JPEG image
        from_image(name, image) — create Image from a PIL.Image.Image object

    Instance methods:
        save(path) — save the image to an external file.
        get_image() — return a PIL.Image.Image object
    """

    classID = 162
    _fields = ScratchMedia._fields + ("form", "rotationCenter", "textBox",
        "jpegBytes", "compositeForm")
    _version = 4

    def built(self):
        # Called after loading from file
        if self.compositeForm:
            self.form_without_text = self.form
            self.form = self.compositeForm


    @classmethod
    def load(cls, path):
        """Load image file and return an Image subclass by format."""
        require_pil()

        (_, name) = os.path.split(path)
        if "." in name:
            name_without_extension = ".".join(name.split(".")[:-1])
        else:
            name_without_extension = name

        image_file = PIL.Image.open(path) # Doesn't read raster data yet :)

        if image_file.format == "JPEG":
            f = open(path, "rb")
            jpegBytes = f.read()
            f.close()

            image = cls(
                name = name_without_extension,
                jpegBytes = ByteArray(jpegBytes),
            )
            image.size = image_file.size
            return image

        else:
            return cls.from_image(name_without_extension, image_file)


    @classmethod
    def from_image(cls, name, image_file):
        """Create Image from a PIL.Image.Image object"""
        if image_file.format == "JPEG":
            f = StringIO.StringIO()
            image_file.save(f, format="JPEG")

            f.seek(0)
            jpegBytes = f.read()

            image = cls(
                name = name,
                jpegBytes = ByteArray(jpegBytes),
            )

        else:
            image_file = image_file.convert("RGBA")
            assert image_file.mode == "RGBA"

            (width, height) = image_file.size
            rgba_string = image_file.tostring()

            image = cls(
                name = name,
                form = Form.from_string(width, height, rgba_string),
            )

        image.size = image_file.size
        return image


    def set_defaults(self):
        ScratchMedia.set_defaults(self)
        self.rotationCenter = Point(0, 0)

        self.form_without_text = None
        self.size = None


    @property
    def width(self):
        (width, height) = self.size
        return width


    @property
    def height(self):
        (width, height) = self.size
        return height


    def get_image(self):
        """Return a PIL.Image.Image object"""
        if self.jpegBytes:
            image = PIL.Image.open(StringIO.StringIO(self.jpegBytes.value))
        else:
            (width, height, rgba_array) = self.form.to_array()
            size = (width, height)
            image = PIL.Image.fromstring("RGBA", size, rgba_array)

        return image


    def save(self, path, format=None):
        """Save the image data to an external file.
        Returns the filename with extension.
        Arguments:
            path - absolute/relative path to save to. Doesn't require extension.
            format - "PNG", "JPEG", etc: passed to PIL.
        """
        guessed_from_extension = False
        (folder, name) = os.path.split(path)
        if "." in name:
            format = name.split('.')[-1]
            format = format.lstrip(".").upper()
            if format == "JPG": format = "JPEG"
            guessed_from_extension = True

        if not format:
            if self.jpegBytes:
                format = "JPEG"
            else:
                format = "PNG"

        if not guessed_from_extension:
            extension = format.lower()
            if extension == "jpeg": extension = "jpg"
            path += "." + extension
            name += "." + extension

        image = self.get_image()
        image.save(path, format)
        return name

    def save_png(self, path):
        """Deprecated. Use .save(path, "PNG") instead"""
        self.save(path, "PNG")

    def save_jpg(self, path):
        """Deprecated. Use .save(path, "JPEG") instead"""
        self.save(path, "JPEG")


class MovieMedia(ScratchMedia):
    """unused?"""
    classID = 163
    _fields = ScratchMedia._fields + ("fileName", "fade", "fadeColor", "zoom",
        "hPan", "vPan", "msecsPerFrame", "currentFrame", "moviePlaying")

class Sound(ScratchMedia):
    classID = 164
    _fields = ScratchMedia._fields + ("originalSound", "volume", "balance",
        "compressedSampleRate", "compressedBitsPerSample", "compressedData")




class KeyEventHatMorph(BaseMorph):
    """unused?"""
    classID = 165

class BooleanArgMorph(BaseMorph):
    """unused?"""
    classID = 166

class EventTitleMorph(BaseMorph):
    """unused?"""
    classID = 167

class MouseClickEventHatMorph(BaseMorph):
    """unused?"""
    classID = 168

class ExpressionArgMorphWithMenu(BaseMorph):
    """unused?"""
    classID = 169

class ReporterBlockMorph(BaseMorph):
    """unused?"""
    classID = 170

class MultilineStringMorph(BorderedMorph):
    """Used for costume text."""
    classID = 171
    _fields = BorderedMorph._fields + ("font", "textColor", "selectionColor",
        "lines")

class ToggleButton(SimpleButtonMorph):
    """unused?"""
    classID = 172

class WatcherReadoutFrameMorph(BorderedMorph):
    """unused?"""
    classID = 173

class WatcherSliderMorph(SimpleSliderMorph):
    """unused?"""
    classID = 174

class ScratchListMorph(BorderedMorph):
    """List of items.
    Attributes:
        name - required
        items
    """
    classID = 175
    _fields = BorderedMorph._fields + ("name", "items", "target")
    _version = 2

    def set_defaults(self):
        BorderedMorph.set_defaults(self)

        self.borderColor = Color(594, 582, 582)
        self.borderWidth = 2
        self.bounds = Rectangle([0, 0, 150, 360]) # ?
        self.color = Color(774, 786, 798)

        self.items = []

    def normalize(self):
        """Called before saving"""
        self.items = [unicode(item) for item in self.items]



class ScrollingStringMorph(BaseMorph):
    """unused"""
    classID = 176







#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2010, David Sveningsson <ext@sidvind.com>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import pprint

f = open('/usr/share/X11/rgb.txt', 'r')
lines = f.readlines()
tokens = [x.split(None, 3) for x in lines]
convert = lambda x: hex(int(x))[2:]
items = [(name[:-1].lower(), (convert(r),convert(g),convert(b))) for (r,g,b,name) in tokens]
d = dict(items)

# Overwrite some colors which differs from x11 rgb.txt
d['gray']   = ('7f', '7f', '7f')
d['green']  = ('00', '7f', '00')
d['maroon'] = ('7f', '00', '00')
d['purple'] = ('7f', '00', '7f')

print '#!/usr/bin/python'
print '# -*- coding: utf-8 -*-'
print
print '# autogenerated from X11/rgb.txt'
print 'names = \\'
pp = pprint.PrettyPrinter(indent=4)
pp.pprint(d)



try:
	import ply
except ImportError, e:
	print "Missing dependency: PLY needed for kurt.scratchblocks module"
	raise e


from lexer import lex, tokens
from parser import block_plugin_parser

def parse_block_plugin(input):
    return block_plugin_parser.parse(input, tracking=True)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2010, David Sveningsson <ext@sidvind.com>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""
Parse HTML/CSS colors
"""

import unittest
import re

from names import names

DecimalFactory = lambda v: int(v,16)
FloatFactory = lambda v: int(v,16)/255.0
HexFactory = lambda v: v

# Get a list of all hexadecimal digits.
_hex_literals = [hex(v)[-1] for v in range(0,16)]

def enforceComponents():
    def decorate(func):
        def wrapper(self, *args, **kwargs):
            n = self.Components
            fill = self.ResultFactory(self.Fill)
            
            assert n in [3,4]
            
            result = func(self, *args, **kwargs)
            components = len(result)
            
            if components == n:
                return result
            elif components > n:
                return result[:n]
            else:
                return result + (n - components) * (fill,)
        return wrapper
    return decorate

class Parser:
    def __init__(self, factory=DecimalFactory, components=3, fill='ff'):
        self.ResultFactory = factory
        self.Components = components
        self.Fill = fill
    
    @enforceComponents()
    def parse(self, string):
        """
        Parses a HTML/CSS color.
        
        :param string: The string to parse
        :return: A tuple containing the color components.
        """
        if not isinstance(string, basestring):
            raise ValueError, 'must be a string'
        
        func = self._detect_format(string)
        result = func(string)
        
        return tuple([self.ResultFactory(x) for x in result])
    
    @staticmethod
    def _parse_hex(string):
        if string[0] == '#':
            string = string[1:]
        
        n = len(string)
        fmt = {
            3: '([0-9A-Fa-f]{1})' * 3, # shorthand RGB
            4: '([0-9A-Fa-f]{1})' * 4, # shorthand RGBA
            6: '([0-9A-Fa-f]{2})' * 3, # RGB
            8: '([0-9A-Fa-f]{2})' * 4 # RGBA
        }
        
        match = re.match(fmt[n], string)
        
        if match:
            groups = match.groups()
            
            # shorthand RGB{,A} must be extended
            if n in [3,4]:
                groups = tuple([2*x for x in groups])
            
            return groups
        else:
            raise ValueError,'Unable to parse "%s"' % (string)
    
    @staticmethod
    def _parse_name(string):
        try:
            return names[string]
        except KeyError:
            raise ValueError,'Unrecognized color name "%s"' % (string)
    
    @staticmethod
    def _detect_format(string):
        if string[0] == '#':
            return Parser._parse_hex
        elif [item for item in string if item not in _hex_literals] == []:
            return Parser._parse_hex
        else:
            return Parser._parse_name

class factory_test(unittest.TestCase):
    def setUp(self):
        self.p = Parser()
    
    def test_decimal(self):
        self.p.ResultFactory = DecimalFactory
        self.p.Components = 3
        self.assertEqual(self.p.parse('#fff'), (255, 255, 255))
    
    def test_float(self):
        self.p.ResultFactory = FloatFactory
        self.p.Components = 3
        self.assertEqual(self.p.parse('#fff'), (1.0, 1.0, 1.0))
    
    def test_hex(self):
        self.p.ResultFactory = HexFactory
        self.p.Components = 3
        self.assertEqual(self.p.parse('#fff'), ('ff', 'ff', 'ff'))
    
    def test_custom(self):
        self.p.ResultFactory = lambda x: int(x,16) * 2
        self.p.Components = 3
        self.assertEqual(self.p.parse('#fff'), (510, 510, 510))

class components_test(unittest.TestCase):
    def setUp(self):
        self.p = Parser()
    
    def test_invalid(self):
        self.p.ResultFactory = DecimalFactory
        for c in [2,5]:
            self.p.Components = c
            self.assertRaises(AssertionError, self.p.parse, '#fff')
    
    def test_valid(self):
        self.p.ResultFactory = DecimalFactory
        self.p.Components = 3
        self.assertEqual(self.p.parse('#fff'), (255, 255, 255))
        self.p.Components = 4
        self.assertEqual(self.p.parse('#ffff'), (255, 255, 255, 255))

class test(unittest.TestCase):
    def setUp(self):
        self.p = Parser(factory=DecimalFactory, components=3)
    
    def test_invalid(self):
        self.assertRaises(ValueError, self.p.parse, 0)
    
    def test_invalid_hex(self):
        self.assertRaises(ValueError, self.p.parse, '#foobar')
    
    def test_short_hex_rgb(self):
        self.assertEqual(self.p.parse('#f70'), (255, 119, 0))
    
    def test_short_hex_rgba(self):
        self.p.Components = 4
        self.assertEqual(self.p.parse('#f70f'), (255, 119, 0, 255))
    
    def test_decimal_rgb(self):
        self.assertEqual(self.p.parse('#ff7700'), (255, 119, 0))
    
    def test_decimal_rgba_fill(self):
        self.p.Components = 4
        self.assertEqual(self.p.parse('#ff7700'), (255, 119, 0, 255))
    
    def test_decimal_rgba(self):
        self.p.Components = 4
        self.assertEqual(self.p.parse('#ff770077'), (255, 119, 0, 119))

    def test_float_rgb(self):
        self.p.ResultFactory = FloatFactory
        [self.assertAlmostEqual(x,y,1) for (x,y) in zip(self.p.parse('#ff7700'), (1.0, 0.46, 0.0))]
    
    def test_float_rgba_fill(self):
        self.p.ResultFactory = FloatFactory
        self.p.Components = 4
        [self.assertAlmostEqual(x,y,1) for (x,y) in zip(self.p.parse('#ff7700'), (1.0, 0.46, 0.0, 1.0))]
    
    def test_float_rgba(self):
        self.p.ResultFactory = FloatFactory
        self.p.Components = 4
        [self.assertAlmostEqual(x,y,1) for (x,y) in zip(self.p.parse('#ff770077'), (1.0, 0.46, 0.0, 0.46))]
    
    def test_sign(self):
        self.assertEqual(self.p.parse('#ff7700'), (255, 119, 0))
        self.assertEqual(self.p.parse('ff7700'), (255, 119, 0))
    
    def test_name(self):
        self.assertEqual(self.p.parse('red'), (255, 0, 0))
    
    def test_invalid_name(self):
        self.assertRaises(ValueError, self.p.parse, 'foobar')



#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Provides information about block types.
Classes:
    BlockType - information about a single type of block.

Values:
    blocks - list of BlockType objects.
    blocks_by_cmd - dict of BlockType objects indexed by their `command`.

Other values:
    block_plugin_inserts - format strings for insert types, used in 
                           Block.to_block_plugin()

The blocks list is compiled by parsing blockspecs copied directly from Scratch's 
Squeak source code.
"""
from construct import *
from construct.text import *

import re

from fixed_objects import Symbol


string = QuotedString("string", start_quote="'", end_quote="'", esc_char="\\")

symbol = Struct("symbol",
    Literal("#"),    
    StringAdapter(GreedyRange(CharOf("value", set("+*/\<>=&|~:-0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"))))
    # _=@%?!`^$
)

spacer = Struct("spacer",
    Literal("#"),
    CharOf("", "~-"),
    Whitespace(),
    Value("is_block", lambda c: False)
)

value = Select("value",
    symbol,
    string,
    FloatNumber("value"),
    DecNumber("value"),
    Struct("negative number",
        Literal("-"),
        DecNumber("number"),
        Value("value", lambda ctx: -ctx['number']),
    ),
        
)

blockspec = Struct("blockspec",
    Literal("("),
    Whitespace(),
    Rename("text", string),
    Whitespace(),
    Rename("flag", symbol),
    Whitespace(),
    Rename("command", symbol),
    Whitespace(),
    OptionalGreedyRepeater(Rename("defaults", Struct("",
        value,
        Whitespace(),
    ))),
    Whitespace(),
    Literal(")"),
    Whitespace(),
    
    Value("is_block", lambda c: True)
)

category = Struct("category",
    Rename("name", string),
    Whitespace(),
    Rename("blocks", OptionalGreedyRepeater(
        Select("",
            blockspec,
            spacer,
        ),
    )),
    Whitespace(),
)

blockspecs = Struct("blockspecs",
    Rename("categories", OptionalGreedyRepeater(category)),
    #StringAdapter(OptionalGreedyRange(Char("leftovers"))), # used for DEBUG
    Terminator,
)



### Taken from Squeak source, with modifications.
# Note: apostrophes ' and slashes \ need special escaping.
squeak_blockspecs = """'control' ('when green flag clicked' #S #EventHatMorph) ('when %k key pressed' #K #KeyEventHatMorph) ('when %m clicked' #M #MouseClickEventHatMorph) #- ('wait %n secs' #t #wait:elapsed:from: 1) #- ('forever' #c #doForever) ('repeat %n' #c #doRepeat 10) #- ('broadcast %e' #- #broadcast:) ('broadcast %e and wait' #s #doBroadcastAndWait) ('when I receive %e' #E #EventHatMorph) #- ('forever if %b' #c #doForeverIf) ('if %b' #c #doIf) ('if %b' #c #doIfElse) ('wait until %b' #s #doWaitUntil) ('repeat until %b' #c #doUntil) #- ('stop script' #s #doReturn) ('stop all' #- #stopAll) 'operators' ('%n + %n' #r #+ #- #-) ('%n - %n' #r #- #- #-) ('%n * %n' #r #* #- #-) ('%n / %n' #r #/ #- #-) #- ('pick random %n to %n' #r #randomFrom:to: 1 10) #- ('%s < %s' #b #< '' '') ('%s = %s' #b #= '' '') ('%s > %s' #b #> '' '') #- ('%b and %b' #b #&) ('%b or %b' #b #|) ('not %b' #b #not) #- ('join %s %s' #r #concatenate:with: 'hello ' 'world') ('letter %n of %s' #r #letter:of: 1 'world') ('length of %s' #r #stringLength: 'world') #- ('%n mod %n' #r #\\\\ #- #-) ('round %n' #r #rounded #-) #- ('%f of %n' #r #computeFunction:of: 'sqrt' 10) ('A block with color %C and color %c' #- #thingwithColor:andColor:) 'sound' ('play sound %S' #- #playSound:) ('play sound %S until done' #s #doPlaySoundAndWait) ('stop all sounds' #- #stopAllSounds) #- ('play drum %D for %n beats' #t #drum:duration:elapsed:from: 48 0.2) ('rest for %n beats' #t #rest:elapsed:from: 0.2) #- ('play note %N for %n beats' #t #noteOn:duration:elapsed:from: 60 0.5) ('set instrument to %I' #- #midiInstrument: 1) #- ('change volume by %n' #- #changeVolumeBy: -10) ('set volume to %n%' #- #setVolumeTo: 100) ('volume' #r #volume) #- ('change tempo by %n' #- #changeTempoBy: 20) ('set tempo to %n bpm' #- #setTempoTo: 60) ('tempo' #r #tempo) 'motor' ('motor on for %n secs' #t #motorOnFor:elapsed:from: 1) ('motor on' #- #allMotorsOn) ('motor off' #- #allMotorsOff) ('motor power %n' #- #startMotorPower: 100) ('motor direction %W' #- #setMotorDirection: 'this way') 'variables' ('show variable %v' #- #showVariable:) ('hide variable %v' #- #hideVariable:) 'list' ('add %s to %L' #- #append:toList: 'thing') #- ('delete %y of %L' #- #deleteLine:ofList: 1) ('insert %s at %i of %L' #- #insert:at:ofList: 'thing' 1) ('replace item %i of %L with %s' #- #setLine:ofList:to: 1 'list' 'thing') #- ('item %i of %L' #r #getLine:ofList: 1) ('length of %L' #r #lineCountOfList:) ('%L contains %s' #b #list:contains: 'list' 'thing')"""

squeak_stage_blockspecs = """'sensing' ('ask %s and wait' #s #doAsk 'What\\'s your name?') ('answer' #r #answer) #- ('mouse x' #r #mouseX) ('mouse y' #r #mouseY) ('mouse down?' #b #mousePressed) #- ('key %k pressed?' #b #keyPressed: 'space') #- ('reset timer' #- #timerReset) ('timer' #r #timer) #- ('%a of %m' #r #getAttribute:of:) #- ('loudness' #r #soundLevel) ('loud?' #b #isLoud) #~ ('%H sensor value' #r #sensor: 'slider') ('sensor %h?' #b #sensorPressed: 'button pressed') 'looks' ('switch to background %l' #- #showBackground: 'background1') ('next background' #- #nextBackground) ('background #' #r #backgroundIndex) #- ('change %g effect by %n' #- #changeGraphicEffect:by: 'color' 25) ('set %g effect to %n' #- #setGraphicEffect:to: 'color' 0) ('clear graphic effects' #- #filterReset) #- 'pen' ('clear' #- #clearPenTrails)"""

squeak_sprite_blockspecs = """'motion' ('move %n steps' #- #forward:) ('turn cw %n degrees' #- #turnRight: 15) ('turn ccw %n degrees' #- #turnLeft: 15) #- ('point in direction %d' #- #heading: 90) ('point towards %m' #- #pointTowards:) #- ('go to x:%n y:%n' #- #gotoX:y: 0 0) ('go to %m' #- #gotoSpriteOrMouse:) ('glide %n secs to x:%n y:%n' #t #glideSecs:toX:y:elapsed:from: 1 50 50) #- ('change x by %n' #- #changeXposBy: 10) ('set x to %n' #- #xpos: 0) ('change y by %n' #- #changeYposBy: 10) ('set y to %n' #- #ypos: 0) #- ('if on edge, bounce' #- #bounceOffEdge) #- ('x position' #r #xpos) ('y position' #r #ypos) ('direction' #r #heading) 'pen' ('clear' #- #clearPenTrails) #- ('pen down' #- #putPenDown) ('pen up' #- #putPenUp) #- ('set pen color to %c' #- #penColor:) ('change pen color by %n' #- #changePenHueBy:) ('set pen color to %n' #- #setPenHueTo: 0) #- ('change pen shade by %n' #- #changePenShadeBy:) ('set pen shade to %n' #- #setPenShadeTo: 50) #- ('change pen size by %n' #- #changePenSizeBy: 1) ('set pen size to %n' #- #penSize: 1) #- ('stamp' #- #stampCostume) 'looks' ('switch to costume %l' #- #lookLike:) ('next costume' #- #nextCostume) ('costume #' #r #costumeIndex) #- ('say %s for %n secs' #t #say:duration:elapsed:from: 'Hello!' 2) ('say %s' #- #say: 'Hello!') ('think %s for %n secs' #t #think:duration:elapsed:from: 'Hmm...' 2) ('think %s' #- #think: 'Hmm...') #- ('change %g effect by %n' #- #changeGraphicEffect:by: 'color' 25) ('set %g effect to %n' #- #setGraphicEffect:to: 'color' 0) ('clear graphic effects' #- #filterReset) #- ('change size by %n' #- #changeSizeBy:) ('set size to %n%' #- #setSizeTo: 100) ('size' #r #scale) #- ('show' #- #show) ('hide' #- #hide) #- ('go to front' #- #comeToFront) ('go back %n layers' #- #goBackByLayers: 1) 'sensing' ('touching %m?' #b #touching:) ('touching color %C?' #b #touchingColor:) ('color %C is touching %C?' #b #color:sees:) #- ('ask %s and wait' #s #doAsk 'What''s your name?') ('answer' #r #answer) #- ('mouse x' #r #mouseX) ('mouse y' #r #mouseY) ('mouse down?' #b #mousePressed) #- ('key %k pressed?' #b #keyPressed: 'space') #- ('distance to %m' #r #distanceTo:) #- ('reset timer' #- #timerReset) ('timer' #r #timer) #- ('%a of %m' #r #getAttribute:of:) #- ('loudness' #r #soundLevel) ('loud?' #b #isLoud) #~ ('%H sensor value' #r #sensor: 'slider') ('sensor %h?' #b #sensorPressed: 'button pressed')"""

squeak_obsolete_blockspecs = """'obsolete number blocks' ('abs %n' #r #abs #-) ('sqrt %n' #r #sqrt #-) 'obsolete sound blocks' ('rewind sound %S' #- #rewindSound:) 'obsolete sprite motion blocks' ('point away from edge' #- #turnAwayFromEdge) ('glide x:%n y:%n in %n secs' #t #gotoX:y:duration:elapsed:from: 50 50 1) 'obsolete sprite looks blocks' ('change costume by %n' #- #changeCostumeIndexBy: 1) ('change background by %n' #- #changeBackgroundIndexBy: 1) #- ('change stretch by %n' #- #changeStretchBy:) ('set stretch to %n%' #- #setStretchTo: 100) #- ('say nothing' #- #sayNothing) #- ('change visibility by %n' #- #changeVisibilityBy: -10) ('set visibility to %n%' #- #setVisibilityTo: 100) 'obsolete image effects' ('change color-effect by %n' #- #changeHueShiftBy: 25) ('set color-effect to %n' #- #setHueShiftTo: 0) #- ('change fisheye by %n' #- #changeFisheyeBy: 10) ('set fisheye to %n' #- #setFisheyeTo: 0) #~ ('change whirl by %n' #- #changeWhirlBy: 30) ('set whirl to %n' #- #setWhirlTo: 0) #- ('change pixelate by %n' #- #changePixelateCountBy: 1) ('set pixelate to %n' #- #setPixelateCountTo: 1) #~ ('change mosaic by %n' #- #changeMosaicCountBy: 1) ('set mosaic to %n' #- #setMosaicCountTo: 1) #- ('change brightness-shift by %n' #- #changeBrightnessShiftBy: 10) ('set brightness-shift to %n' #- #setBrightnessShiftTo: 0) #~ ('change saturation-shift by %n' #- #changeSaturationShiftBy: 10) ('set saturation-shift to %n' #- #setSaturationShiftTo: 0) #- ('change pointillize drop by %n' #- #changePointillizeSizeBy: 5) ('set pointillize drop to %n' #- #setPointillizeSizeTo: 0) #~ ('change water ripple by %n' #- #changeWaterRippleBy: 5) ('set water ripple to %n' #- #setWaterRippleTo: 0) #- ('change blur by %n' #- #changeBlurBy: 1) ('set blur to %n' #- #setBlurTo: 0)"""
### End Squeak code



class BlockType:
    """Information about a single type of block.
    Attributes:
        command - the command used in Squeak to run the block. (see Block.name)
        text - text that appears on the block. 
               Contains inserts starting with % signs.
        parts - text, split up into text segments and inserts.
        flag - a single char describing the kind of block.
        category - where this block is found in Scratch's interface.
        defaults - list of default values for block inserts. (see Block.args)
    """
    INSERT_RE = re.compile(r'(%.)')
    
    def __init__(self, command, text, flag='-', category='', defaults=None):
        self.command = command
        self.text = text
        self.flag = flag
        self.category = category
        if defaults is None: defaults = []
        self.defaults = defaults
    
    def copy(self):
        return BlockType(
            self.command, self.text, self.flag, self.category, self.defaults[:]
        )
    
    @property
    def parts(self):
        return self.INSERT_RE.split(self.text)
    
    @property
    def inserts(self):
        return filter(lambda p: p and len(p) == 2 and p[0] == "%", self.parts)
    
    def __repr__(self):
        return '<BlockType(%s)>' % self.command
    
    def make_default(self, script=None):
        """Returns a new Block object of this type with default arguments."""
        return Block(self.command, *self.defaults[:])


def parse_blockspec(squeak_code):
    parsed = blockspecs.parse(squeak_code)
    categories = parsed.categories
    
    blocks = []
    for category in categories:
        for block in category.blocks:
            if not block.is_block:
                continue
            
            defaults = []
            for default in block.defaults:
                default = default.value
                if isinstance(default, Container):
                    default = default.value
                defaults.append(default)                        
            
            block = BlockType(
                block.command.value,
                block.text,
                block.flag.value,
                category.name, 
                defaults,
            )
            blocks.append(block)
    
    return blocks



blocks = (list(parse_blockspec(squeak_blockspecs)) + 
    list(parse_blockspec(squeak_stage_blockspecs)) + 
    list(parse_blockspec(squeak_sprite_blockspecs)) + 
    list(parse_blockspec(squeak_obsolete_blockspecs)))

blocks += [
    BlockType("readVariable", "%v", "r", category="variables"),
    BlockType("changeVariable", "change %v by %n", category="variables", 
        defaults = [None, Symbol("changeVar:by:"), None]),
    BlockType("changeVariable", "set %v to %s", category="variables",
        defaults = [None, Symbol("setVar:to:"), None]),
    
    BlockType("contentsOfList:", "%l", "r", "variables"),
    
    BlockType("EventHatMorph", "when gf clicked", "S",  # alternate spelling
        defaults = ["Scratch-StartClicked"]),
    
    BlockType("", "obsolete!"),
]

blocks_by_cmd = {}
for block in blocks:
    cmd = block.command
    if cmd not in blocks_by_cmd:
        blocks_by_cmd[cmd] = []
    blocks_by_cmd[cmd].append(block)


# Fixes
assert blocks_by_cmd['EventHatMorph'][0].text == 'when green flag clicked'
blocks_by_cmd['EventHatMorph'][0].defaults = ["Scratch-StartClicked"]

assert blocks_by_cmd['MouseClickEventHatMorph'][0].text == 'when %m clicked'
blocks_by_cmd['MouseClickEventHatMorph'][0].defaults = ["Scratch-MouseClickEvent"]


def strip_block_text(parts):
    """Returns text with spaces and inserts removed."""
    text = ''.join(part.replace(" ", "") for part in parts
                    if part and not (len(part) == 2 and part[0] == "%"))
    for chr in "?:-":
        new_text = text.replace(chr, "")
        if new_text:
            text = new_text
    return text.lower()


blocks_by_text = {}
for block in blocks:
    text = strip_block_text(block.parts)
    
    # Some blocks have same text
    if text not in blocks_by_text:
        blocks_by_text[text] = []
    blocks_by_text[text].append(block)


def find_block(text):
    parts = BlockType.INSERT_RE.split(text)
    text = strip_block_text(parts)
    if text in blocks_by_text:
        return blocks_by_text[text]
    else:
        return []


block_plugin_inserts = {
    "%b": "<%s>",
    "%n": "(%s)",
    "%d": "(%s v)",
    "%s": "[%s]",
    "%c": "[#%s]",
    "%C": "[#%s]",
        
    "%m": "[%s v]",
    "%a": "[%s v]",
    "%e": "[%s v]",
    "%k": "[%s v]",
        
    "%v": "[%s v]",
    "%L": "[%s v]",
    "%i": "(%s v)",
    "%y": "(%s v)",
        
    "%f": "[%s v]",
        
    "%l": "[%s v]",
    "%g": "[%s v]",
        
    "%S": "[%s v]",
    "%D": "(%s v)",
    "%N": "(%s v)",
    "%I": "(%s v)",
    
    "%h": "[%s v]",
    "%H": "[%s v]",
    "%W": "[%s v]",
}


from scripts import Block




#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Primitive fixed-format objects - eg String, Dictionary."""

from construct import Container, Struct, Embed, Rename
from construct import PascalString, UBInt32, SBInt32, UBInt16, UBInt8, Bytes
from construct import BitStruct, Padding, Bits
from construct import Value, Switch, If, IfThenElse, OptionalGreedyRepeater
from construct import Array as StrictRepeater, Array as MetaRepeater
# We can't import the name Array, as we use it. -_-
import construct

# used by Form
from array import array
try:
    import png
except ImportError:
    png = None

from inline_objects import Field


def default_colormap():
    # This generation comes from the squeak source under initializeIndexColors
    colormap = []
    # 1-bit colors
    colormap.append((255, 255, 255, 255))
    colormap.append((0, 0, 0, 255))
    # additional 2-bit colors
    colormap.append((255, 255, 255, 255))
    colormap.append((127, 127, 127, 255))
    # additional 4-bit colors
    colormap.append((255, 0, 0, 255))
    colormap.append((0, 255, 0, 255))
    colormap.append((0, 0, 255, 255))
    colormap.append((0, 255, 255, 255))
    colormap.append((255, 255, 0, 255))
    colormap.append((255, 0, 255, 255))
    colormap.append((31, 31, 31, 255))
    colormap.append((63, 63, 63, 255))
    colormap.append((95, 95, 95, 255))
    colormap.append((159, 159, 159, 255))
    colormap.append((191, 191, 191, 255))
    colormap.append((223, 223, 223, 255))
    # additional 8-bit colors
    for i in range(32):  # 24 more shades of gray
        if i % 4 == 0:
            continue
        value = 255 * i / 32
        colormap.append((value, value, value, 255))
    for red in range(6):  # Color "cube" with six steps for each primary color
        for blue in range(6):
            for green in range(6):
                colormap.append((255 * red / 5,
                                 255 * green / 5,
                                 255 * blue / 5,
                                 255))
    return colormap


class FixedObject(object):
    """A primitive fixed-format object - eg String, Dictionary.
    value property - contains the object's value."""
    def __init__(self, value):
        if isinstance(value, FixedObject):
            value = value.value
        self.value = value
    
    def to_construct(self, context):
        return Container(
            classID = self.__class__.__name__, 
            value = self.to_value(),
        )
    
    @classmethod
    def from_construct(cls, obj, context):
        fixed_obj = cls.from_value(obj.value)
        return fixed_obj
    
    def to_value(self):
        return self.value
    
    @classmethod
    def from_value(cls, value):
        return cls(value)
    
    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.value == other.value
    
    def __ne__(self, other):
        return not self == other
    
    def __str__(self):
        return repr(self)
    
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.value)
    
    def copy(self):
        return self.__class__(self.value)


class ContainsRefs: pass

class FixedObjectWithRepeater(FixedObject):
    """Used internally to handle things like
        Struct("",
            UBInt32("length"),
            MetaRepeater(lambda ctx: ctx.length, UBInt32("items")),
        )
    """
    def to_value(self):
        return Container(items = self.value, length = len(self.value))
    
    @classmethod
    def from_value(cls, obj):
        assert len(obj.items) == obj.length, "File corrupt?"
        return cls(obj.items)


class FixedObjectByteArray(FixedObject):
    def __repr__(self):
        name = self.__class__.__name__
        value = repr(self.value)
        if len(value) > 60:
            value = value[:97] + '...'
            return "<%s(%s)>" % (name, value)
        else:
            return "%s(%s)" % (name, value)

# Bytes

class String(FixedObjectByteArray):
    classID = 9
    _construct = PascalString("value", length_field=UBInt32("length"))


class Symbol(FixedObjectByteArray):
    classID = 10
    _construct = PascalString("value", length_field=UBInt32("length"))
    def __repr__(self):
        return "<#%s>" % self.value


class ByteArray(FixedObjectByteArray):
    classID = 11
    _construct = PascalString("value", length_field=UBInt32("length"))
    
    def __repr__(self):
        return '<%s(%i bytes)>' % (self.__class__.__name__, len(self.value))


class SoundBuffer(FixedObjectByteArray, FixedObjectWithRepeater):    
    classID = 12
    _construct = Struct("",
        UBInt32("length"),
        MetaRepeater(lambda ctx: ctx.length, UBInt16("items")),
    )

# Bitmap 13 - found later in file

class UTF8(FixedObjectByteArray):
    classID = 14
    _construct = PascalString("value", length_field=UBInt32("length"), 
                              encoding="utf8")




# Collections

class Collection(FixedObjectWithRepeater, ContainsRefs):
    _construct = Struct("",
        UBInt32("length"),
        MetaRepeater(lambda ctx: ctx.length, Rename("items", Field)),
    )
    
    def __init__(self, value=None):
        if value == None:
            value = []
        FixedObject.__init__(self, value)
    
    def __iter__(self):
        return iter(self.value)
    
    def __getattr__(self, name):
        if name in ('append', 'count', 'extend', 'index', 'insert', 'pop', 
                    'remove', 'reverse', 'sort'):
            return getattr(self.value, name)
    
    def __getitem__(self, index):
        return self.value[index]
    
    def __setitem__(self, index, value):
        self.value[index] = value
    
    def __delitem__(self, index):
        del self.value[index]
    
    def __len__(self):
        return len(self.value)
    
    def copy(self):
        return self.__class__(list(self.value.copy))

class Array(Collection):
    classID = 20
class OrderedCollection(Collection):
    classID = 21
class Set(Collection):
    classID = 22
class IdentitySet(Collection):
    classID = 23



# Dictionary

class Dictionary(Collection):
    classID = 24
    _construct = Struct("dictionary",
        UBInt32("length"),
        MetaRepeater(lambda ctx: ctx.length, Struct("items",
            Rename("key", Field),
            Rename("value", Field),
        )),
    )
    
    def __init__(self, value=None):
        if value == None: value = {}
        Collection.__init__(self, value)
    
    def to_value(self):
        items = [Container(key=key, value=value) 
                 for (key, value) in dict(self.value).items()]
        return Container(items=items, length=len(items))
    
    @classmethod
    def from_value(cls, obj):
        value = dict([(item.key, item.value) for item in obj.items])
        return cls(value)
    
    def __getattr__(self, name):
        return getattr(self.value, name)
    
    def copy(self):
        return self.__class__(self.value.copy())

class IdentityDictionary(Dictionary):
    classID = 25



# Color

class Color(FixedObject):
    """A 32-bit RGB color value.
    Each component r, g, b has a value between 0 and 1023.
    
    However, Colors are considered equal if they have the same 8-bit value.
    """
    classID = 30
    _construct = BitStruct("value",
        Padding(2),
        Bits("r", 10),
        Bits("g", 10),
        Bits("b", 10),
    )
    
    _construct_32_rgba = Struct("",
        UBInt8("r"),
        UBInt8("g"),
        UBInt8("b"),
        UBInt8("alpha"),
    )
    
    def __init__(self, r, g, b):
        self.r = r
        self.g = g
        self.b = b
    
    def __eq__(self, other):
        return (
            isinstance(other, Color) and
            self.to_8bit() == other.to_8bit()
        )
    
    def __ne__(self, other):
        return not self == other
    
    def to_value(self):
        return Container(r=self.r, g=self.g, b=self.b)
    
    @classmethod
    def from_value(cls, value):
        return cls(value.r, value.g, value.b)
    
    @classmethod
    def from_8bit(self, r, g=None, b=None):
        if g is None and b is None:
            rgb = r
        else:
            rgb = (r, g, b)
        
        return Color(*(x << 2 for x in rgb))
        

    @property
    def value(self):
        return (self.r, self.g, self.b)
    
    def copy(self):
        return Color(self.r, self.g, self.b)
    
    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            repr(self.value).strip("()"),
        )
    
    def to_8bit(self):
        """Returns value with components between 0-255."""
        return tuple(x >> 2 for x in self.value)
    
    def to_rgba_array(self):
        (r, g, b) = self.to_8bit()
        return array('B', (r, g, b, 255))
    
    def hexcode(self):
        """Returns the color value in hex/HTML format.
        eg "ff1056".
        """
        hexcode = ""
        for x in self.to_8bit():
            part = hex(x)[2:]
            if len(part) < 2: part = "0" + part
            hexcode += part
        return hexcode


class TranslucentColor(Color):
    classID = 31
    _construct = Struct("",
        Embed(Color._construct),
        UBInt8("alpha"), # I think.
    )
    _construct_32 = Struct("",
        UBInt8("alpha"),
        UBInt8("r"),
        UBInt8("g"),
        UBInt8("b"),
    )

    def __init__(self, r, g, b, alpha):
        self.r = r
        self.g = g
        self.b = b
        self.alpha = alpha
    
    def to_value(self):
        return Container(r=self.r, g=self.g, b=self.b, alpha=self.alpha)
    
    @classmethod
    def from_value(cls, value):
        return cls(value.r, value.g, value.b, value.alpha)
    
    def copy(self):
        return TranslucentColor(self.r, self.g, self.b, self.alpha)
    
    @classmethod
    def from_32bit_raw_argb(cls, raw):
        container = cls._construct_32.parse(raw)
        parts = cls.from_value(container)
        color = cls(*(x << 2 for x in parts.value))
        if color.alpha == 0 and (color.r > 0 or color.g > 0 or color.b > 0):
            color.alpha = 1023
        return color
    
    def to_rgba_array(self):
        return array('B', self.to_8bit())
    
    @property
    def value(self):
        return (self.r, self.g, self.b, self.alpha)

    def hexcode(self, include_alpha=True):
        """Returns the color value in hex/HTML format.
        eg "ff1056ff".
        Argument include_alpha: default True.
        """
        hexcode = Color.hexcode(self)
        if not include_alpha: hexcode = hexcode[:-2]
        return hexcode



# Dimensions

class Point(FixedObject):  
    classID = 32
    _construct = Struct("",
        Rename("x", Field),
        Rename("y", Field),
    )
    
    def __init__(self, x, y=None):
        if y is None: (x, y) = x
        self.x = x
        self.y = y
    
    @property
    def value(self):
        return (self.x, self.y)
    
    def __iter__(self):
        return iter(self.value)
    
    def __repr__(self):
        return 'Point(%r, %r)' % self.value
    
    def to_value(self):
        return Container(x = self.x, y = self.y)
    
    @classmethod
    def from_value(cls, value):
        return cls(value.x, value.y)
    
    def copy(self):
        return Point(self.x, self.y)
    
    @classmethod
    def from_string(cls, string):
        (x, y) = string.split(",")
        return cls(float(x), float(y))

class Rectangle(FixedObject):
    classID = 33
    _construct = StrictRepeater(4, Field)
    
    @classmethod
    def from_value(cls, value):
        value = list(value)
        return cls(value)




# Form/images

def get_run_length(ctx):
    try:
        return ctx.run_length
    except AttributeError:
        return ctx._.run_length

class Bitmap(FixedObjectByteArray, FixedObjectWithRepeater):
    classID = 13
    _construct = Struct("",
        UBInt32("length"),
        construct.String("items", lambda ctx: ctx.length * 4),
        # Identically named "String" class -_-
    )
        
    @classmethod
    def from_value(cls, obj):
        return cls(obj.items)

    def to_value(self):
        value = self.value
        length = (len(value) + 3) / 4
        value += "\x00" *  (length * 4  -  len(value)) # padding
        return Container(items = value, length = length)
    
    _int = Struct("int",
        UBInt8("_value"),
        If(lambda ctx: ctx._value > 223,
            IfThenElse("", lambda ctx: ctx._value <= 254, Embed(Struct("",
                UBInt8("_second_byte"),
                Value("_value", 
                    lambda ctx: (ctx._value - 224) * 256 + ctx._second_byte),
            )), Embed(Struct("",
                UBInt32("_value"),
            )))
        ),
    )
    
    _length_run_coding = Struct("",
        Embed(_int), #ERROR?
        Value("length", lambda ctx: ctx._value),
        
        OptionalGreedyRepeater(
            Struct("data",
                Embed(_int),
                Value("data_code", lambda ctx: ctx._value % 4),
                Value("run_length", lambda ctx:
                    (ctx._value - ctx.data_code) / 4),
                Switch("", lambda ctx: ctx.data_code, {
                    0: Embed(Struct("",
                        StrictRepeater(get_run_length,
                            Value("pixels", lambda ctx: "\x00\x00\x00\x00")
                        ),
                    )),
                    1: Embed(Struct("",
                        Bytes("_b", 1),
                        StrictRepeater(get_run_length,
                            Value("pixels", lambda ctx: ctx._b * 4),
                        ),
                    )),
                    2: Embed(Struct("",
                        Bytes("_pixel", 4),
                        StrictRepeater(get_run_length,
                            Value("pixels", lambda ctx: ctx._pixel),
                        ),
                    )),
                    3: Embed(Struct("",
                        StrictRepeater(get_run_length,
                            Bytes("pixels", 4),
                        ),
                    )),
                }),
            )
        )
    )
    
    @classmethod
    def from_byte_array(cls, bytes):
        """Decodes a run-length encoded ByteArray and returns a Bitmap.
        The ByteArray decompresses to a sequence of 32-bit values, which are
        stored as a byte string. (The specific encoding depends on Form.depth.)
        """
        runs = cls._length_run_coding.parse(bytes)
        data = "" 
        for run in runs.data:
            for pixel in run.pixels:
                data += pixel
        return cls(data)
    
    
    def compress(self):
        """Compress to a ByteArray"""
        raise NotImplementedError



class Form(FixedObject, ContainsRefs):
    """A rectangular array of pixels, used for holding images.
    Attributes:
        width, height - dimensions
        depth - how many bits are used to specify the color at each pixel.
        bits - a Bitmap with varying internal structure, depending on depth.
        privateOffset - ?
    
    Note: do not modify the dict returned from the .value property.
    """
    
    classID = 34
    _construct = Struct("form",
        Rename("width", Field),
        Rename("height", Field),
        Rename("depth", Field),
        Rename("privateOffset", Field),
        Rename("bits", Field), # Bitmap
    )
    
    def __init__(self, **fields):
        self.width = 0
        self.height = 0
        self.depth = None
        self.privateOffset = None
        self.bits = Bitmap("")
        self.colors = None
        
        self.__dict__.update(fields)
    
    @property
    def value(self):
        return dict((k, getattr(self, k)) for k in self.__dict__ 
                    if not k.startswith("_"))
    
    def to_value(self):
        return Container(**self.value)
    
    @classmethod
    def from_value(cls, value):
        return cls(**dict(value))
    
    def copy(self):
        return self.__class__.from_value(self.to_value())
    
    def __repr__(self):
        return "<%s(%ix%i)>" % (
            self.__class__.__name__,
            self.width, self.height,
        )
    
    def built(self):
        if isinstance(self.bits, ByteArray):
            self.bits = Bitmap.from_byte_array(self.bits.value)
        assert isinstance(self.bits, Bitmap)
    
    def _to_pixels(self):
        pixel_bytes = self.bits.value
        if self.depth == 32:
            for i in range(0, len(pixel_bytes), 4):
                (a, r, g, b) = (ord(x) for x in pixel_bytes[i:i+4])
                if a == 0 and (r > 0 or g > 0 or b > 0):
                    a = 255
                yield array("B", (r, g, b, a))
        
        elif self.depth == 16:
            raise NotImplementedError # TODO: depth 16
        
        elif self.depth <= 8:
            if self.colors:
                colors = [color.to_rgba_array() for color in self.colors]
            else:
                colors = default_colormap()
            length = len(pixel_bytes) * 8 / self.depth
            repeater = MetaRepeater(length, Bits("pixels", self.depth))
            pixels_construct = BitStruct("", repeater)
            for pixel in pixels_construct.parse(pixel_bytes).pixels:
                yield colors[pixel]
    
    def to_array(self):
        rgba = array('B') #unsigned byte
        pixel_count = 0
        num_pixels = self.width * self.height

        # Rows are rounded to be a whole number of words (32 bits) long.
        # Presumably this is because Bitmaps are compressed (run-length encoded)
        # in 32-bit segments.
        skip = 0
        if self.depth <= 8:
            pixels_per_word = 32 / self.depth
            pixels_in_last_word = self.width % pixels_per_word
            skip = (pixels_per_word - pixels_in_last_word) % pixels_per_word
        
        x = 0
        pixels = self._to_pixels()
        while 1:
            try:
                color = pixels.next()
            except StopIteration:
                break
            
            rgba.extend(color)
            
            pixel_count += 1
            x += 1            
            if x >= self.width:
                for i in xrange(skip):
                    pixel = pixels.next()
                x = 0
        
        length = self.width * self.height * 4
        blank = array("B", (0, 0, 0, 0))
        while len(rgba) < length: 
            rgba.extend(blank)
        
        assert len(rgba) == length
        
        return (self.width, self.height, rgba)

    @classmethod
    def from_string(cls, width, height, rgba_string):
        """Returns a Form with 32-bit RGBA pixels
        Accepts string containing raw RGBA color values
        """
        # Convert RGBA string to ARGB
        raw = ""
        for i in range(0, len(rgba_string), 4):
            raw += rgba_string[i+3]   # alpha
            raw += rgba_string[i:i+3] # rgb
        
        assert len(rgba_string) == width * height * 4
        
        return Form(
            width = width,
            height = height,
            depth = 32,
            bits = Bitmap(raw),
        )

    @classmethod
    def from_array(cls, width, height, rgba_array):
        """Returns a Form with 32-bit RGBA pixels
        Accepts sequence of flattened r, g, b, a, r, g, b, a ... values for 
        each pixel
        """
        # Unused now
        raw = ""
        for i in range(0, len(rgba_array), 4):
            (r, g, b, a) = (chr(x) for x in rgba_array[i:i+4])
            raw += "".join((a, r, g, b))

        return Form(
            width = width,
            height = height,
            depth = 32,
            bits = Bitmap(raw),
        )


class ColorForm(Form):
    """A rectangular array of pixels, used for holding images.
    width, height - dimensions
    depth - how many bits are used to specify the color at each pixel.
    bits - a Bitmap with varying internal structure, depending on depth.
    colors - the colors pointed to by the bits array. (I think?)
    privateOffset - ?
    """
    classID = 35
    _construct = Struct("",
        Embed(Form._construct),
        Rename("colors", Field), # Array
    )


        







#!/usr/bin/python
#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Graphical interface for kurtc.py"""

from Tkinter import *
from tkFont import Font
import tkFileDialog
import tkMessageBox

import subprocess
import sys
import os
import shutil

try:
    import kurt
except ImportError: # try and find kurt directory
    path_to_file = os.path.join(os.getcwd(), __file__)
    path_to_lib = os.path.split(os.path.split(path_to_file)[0])[0]
    sys.path.append(path_to_lib)

from kurt.files import *
from kurt.compiler import *
from kurt.decompiler import *



# Attempt to find My Documents / home folder
HOME_FOLDER = os.path.expanduser("~")
try:
    import win32com.client
except ImportError:
    win32com = None
if win32com:
    objShell = win32com.client.Dispatch("WScript.Shell")
    HOME_FOLDER = objShell.SpecialFolders("MyDocuments")



def open_file(path):
    """Opens Explorer/Finder with given path, depending on platform"""
    if sys.platform=='win32':
        os.startfile(path)
        #subprocess.Popen(['start', path], shell= True)
    
    elif sys.platform=='darwin':
        subprocess.Popen(['open', path])
    
    else:
        try:
            subprocess.Popen(['xdg-open', path])
        except OSError:
            pass
            # xdg-open *should* be supported by recent Gnome, KDE, Xfce


class App:
    def __init__(self, master):
        self.master = master
        content = Frame(master)
        content.grid(column=0, row=0)
        
        # File chooser
        self.file_path = Entry(content, width=80)
        self.file_path.grid(column=0, row=0, columnspan=2, sticky="NSEW", padx=4, pady=4)
        Button(content,
            text = "Choose .sb file",
            command = self.choose_file,
        ).grid(column=0, row=1, sticky="E", padx=8, pady=4)
        Button(content,
            text = "Choose directory",
            command = self.choose_folder,
        ).grid(column=1, row=1, sticky="W", padx=8, pady=4)
        
        # hr
        Frame(content, height=4, bd=2, relief="sunken").grid(column=0, row=2, columnspan=2, sticky="NSEW", pady=8)
        
        # Action buttons
        self.action_font = Font(size=16)
        Button(content,
            text = "Decompile",
            command = self.decompile,
            font = self.action_font,
        ).grid(column=0, row=3, sticky="E", padx=8, pady=4)
        Button(content,
            text = "Compile",
            command = self.compile,
            font = self.action_font,
        ).grid(column=1, row=3, sticky="WN", padx=8, pady=4)
         
        dl = Message(content, width=300, text="""scratch project -> folder structure

Export project contents to separate files. Creates a folder called "<project name> files", with a subfolder for each sprite.
        """)
        dl.grid(column=0, row=4, sticky="E", padx=8)
        cl = Message(content, width=300, text="""folder structure -> scratch project

Take a folder structure generated by decompile, and compiles all the images and scripts back into a .sb file.
        """)
        cl.grid(column=1, row=4, sticky="WN", padx=8)
        
        self.set_file_path(HOME_FOLDER)
        self.file_path.focus_set()
    
    def choose_file(self):
        filename = tkFileDialog.askopenfilename(
            title = "Choose Scratch project",
            initialfile = self.get_file(),
            defaultextension = 'sb',
            filetypes = ("{Scratch Project Files} {.sb}",)
        )
        if filename:
            self.set_file(filename)
            
    def choose_folder(self):
        print self.get_folder()
        path = tkFileDialog.askdirectory(
            initialdir = self.get_folder() or HOME_FOLDER,
            title = "Choose folder containing project files",
        )
        
        if path:
            self.set_folder(path)
    
    def get_folder(self):
        """Return "<project> files" folder"""
        path = self.file_path.get()
        if not path:
            return
        
        # Work out "<project> files" folder from .sb file
        # TODO
        
        return path
        
    def get_file(self):
        """Return selected .sb file"""
        path = self.file_path.get()
        if not path:
            return
        
        # Work out .sb file from "<project> files" folder
        # TODO
        
        return path
    
    def set_file(self, filename):
        """Set selection to .sb file"""
        self.set_file_path(filename)
    
    def set_folder(self, path):
        """Set selection to "<project> files" folder"""
        self.set_file_path(path)
    
    def set_file_path(self, path):
        """Update the file_path Entry widget"""
        self.file_path.delete(0, END)
        self.file_path.insert(0, path)
    
    def decompile(self):
        path = self.get_file()
        if not path:
            tkMessageBox.showerror("Invalid path", "Enter path to .sb file")
            return
        
        project = ScratchProjectFile(path, load=False)
        error = None
        
        try:
            decompile(project)
        except FolderExistsException, e:
            path = e.message
            response = tkMessageBox.askokcancel("Folder Exists", "WARNING: Overwrite folder '%s'?" % path)
            if response:
                shutil.rmtree(path)
                self.decompile()
            return
        
        except FileNotFoundException, e:
            error = "Missing file: %s" % unicode(e)
        except InvalidProject, e:
            error = "Invalid project: %s" % unicode(e)
        
        if error:
            tkMessageBox.showerror("Error", error)
        else:        
            (project_dir, name) = os.path.split(project.path)
            folder = os.path.join(project_dir, "%s files" % project.name)
        
            self.set_folder(folder)
            tkMessageBox.showinfo("Decompiled!", "Saved in folder %s" % folder)
            open_file(folder)
    
    def compile(self):
        path = self.get_file()
        if not path:
            tkMessageBox.showerror("Invalid path", 'Enter path to "<project> files" folder')
            return
        
        if path.endswith(".sb"):
            path = path[:-3]
        if path.endswith(" files"):
            path = path[:-6]
        
        error = None
        try:
            project = compile(path)
            project.save()
        
        except FileExistsException, e:
            path = e.message
            response = tkMessageBox.askokcancel("File exists", "WARNING: Overwrite file '%s'? THIS CANNOT BE UNDONE." % path)
            if response:
                os.remove(path)
                self.compile()
            return
        
        except InvalidFile, e:
            error = "Invalid file: %s" % e
        except FolderNotFoundException, e:
            error = "Missing folder: %s" % unicode(e)
        
        if error:
            tkMessageBox.showerror("Error", error)
        else:
            file = project.path
            self.set_file(file)
            tkMessageBox.showinfo("Compiled!", "Saved to %s" % file)
            open_file(file)
    
    def closed(self):
        pass # Save, perhaps?



root = Tk()
root.title("kurt compiler")
root.resizable(0, 0)

app = App(root)


def closed():
    app.closed()
    root.destroy()

root.protocol('WM_DELETE_WINDOW', closed)


root.mainloop()


#!/usr/bin/python
#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Scratch project  <--->  folder structure of scripts and images

Usage: kurtc.py decompile "path/to/file.sb"
       kurtc.py compile "path/to/project files/"

decompile: scratch project  -> folder structure with project contents
compile:   folder structure -> scratch project

Folder structure: (most subfolders are optional)
    00 Stage/                 [each sprite has its own directory]
        backgrounds/
        ...
    01 Sprite 1/
        costumes/             [export to PNG or JPG format files, 
            01 costume1.png              import from most formats]
            02 costume2.jpg
            ...
        costumes.txt          [costume details, rotation centers]
        lists/                [.txt file for each list, one item per line]
        scripts/              [.txt files: block plugin (scratchblocks) syntax]
            01 when green flag clicked.txt
            ...
        variables.txt         [variable = value, one per line]"""

import os
import sys

try:
    import kurt
except ImportError: # try and find kurt directory
    path_to_file = os.path.join(os.getcwd(), __file__)
    path_to_lib = os.path.split(os.path.split(path_to_file)[0])[0]
    sys.path.append(path_to_lib)

from kurt.compiler import cmd_compile
from kurt.decompiler import cmd_decompile



if __name__ == '__main__':
    command = None
    if len(sys.argv) > 1:
        command = sys.argv[1]
        
        path = None
        if len(sys.argv) > 2:
            path = sys.argv[2]
    
    if not command:
        print __doc__
    
    else:
        if command == "decompile":
            cmd_f = cmd_decompile
        elif command == "compile":
            cmd_f = cmd_compile
        else:
            if not command or command == "help":
                print __doc__
            else:
                print "Invalid command:", command
            cmd_f = None
        
        if cmd_f:
            if path:
                project = cmd_f(path)
            else:
                print cmd_f.__doc__



#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Inline values - such as nil, boolean, int - and references."""

from construct import *
from construct import Array as MetaRepeater
# We can't import the name Array, as we use it. -_-


### inline fields & Refs ###

class Ref(object):
    """A reference to an object located in a serialized object table.
    Used internally for parsing obj_tables.
    Allows for complex interlinked object networks to be serialized to a list of 
    objects.
    Found in UserObjects and certain FixedObjects.
    """
    def __init__(self, index):
        """Initialise a reference.
        @param index: the index in the object table that the reference points to
        Note that the first index is 1.
        """
        self.index = int(index)
    
    def to_construct(self):
        #index1 = self.index % 65536
        #index2 = (self.index - index1) >> 16
        #return Container(classID = 'Ref', _index1 = index1, _index2 = index2)
        return Container(classID="Ref", index=self.index)
    
    @classmethod
    def from_construct(cls, obj):
        index = obj.index #int(obj._index2 << 16) + obj._index1
        return Ref(index)
    
    def __repr__(self):
        return 'Ref(%i)' % self.index
    
    def __eq__(self, other):
        return isinstance(other, Ref) and self.index == other.index
    
    def __ne__(self, other):
        return not self == other
    
    def __hash__(self):
        return hash(self.index)


class RefAdapter(Adapter):
    def _encode(self, obj, context):
        assert isinstance(obj, Ref)
        return obj.to_construct()
        
    def _decode(self, obj, context):
        return Ref.from_construct(obj)


class LargeIntegerAdapter(Adapter):
    def __init__(self, sign, *args, **kwargs):
        self.sign = sign
        Adapter.__init__(self, *args, **kwargs)
    
    def _decode(self, obj, context):
        value = 0
        for x in reversed(obj.data):
            value <<= 8
            value += x
        if self.sign == '-':
            value = -value
        return value
        
    def _encode(self, obj, context):
        bytes = []
        while obj:
            bytes.append(obj & 0xff)
            obj >>= 8
        return Container(length=len(bytes), data=bytes)


class FieldAdapter(Adapter):
    def _encode(self, obj, context):
        assert not isinstance(obj, str)
        
        if obj is None:
            classID = 'nil'
        elif obj is True:
            classID = 'true'
        elif obj is False:
            classID = 'false'
        elif isinstance(obj, float):
            classID = 'Float'
        elif isinstance(obj, Ref):
            classID = 'Ref'
        elif isinstance(obj, int) or isinstance(obj, long):
            if -32768 <= obj <= 32767:
                classID = 'SmallInteger16'
            elif -2147483648 <= obj <= 2147483647:
                classID = 'SmallInteger'
            elif obj >= 0:
                classID = 'LargePositiveInteger'
            else:
                classID = 'LargeNegativeInteger'
        else:
            raise NotImplementedError, 'no field type for %r' % obj
        return Container(classID=classID, value=obj)
    
    def _decode(self, obj, context):
        if isinstance(obj, Ref):
            return obj
        else:
            return obj.value


Field = FieldAdapter(Struct("field",
    Enum(UBInt8("classID"),
        nil = 1,
        true = 2,
        false = 3,
        SmallInteger = 4,
        SmallInteger16 = 5,
        LargePositiveInteger = 6,
        LargeNegativeInteger = 7,
        Float = 8,
        Ref = 99,
    ),
    Switch("value", lambda ctx: ctx.classID, {
        "nil": Value("", lambda ctx: None),
        "true": Value("", lambda ctx: True),
        "false": Value("", lambda ctx: False),
        "SmallInteger": SBInt32(""),
        "SmallInteger16": SBInt16(""),
        "LargePositiveInteger": LargeIntegerAdapter('+', Struct("",
            UBInt16("length"),
            MetaRepeater(lambda ctx: ctx.length, UBInt8("data")),
        )),
        "LargeNegativeInteger": LargeIntegerAdapter('-', Struct("",
            UBInt16("length"),
            MetaRepeater(lambda ctx: ctx.length, UBInt8("data")),
        )),
        "Float": BFloat64(""),
        "Ref": RefAdapter(BitStruct("",
            BitField("index", 24),
        )),
    })
))
Field.__doc__ = """Construct for simple inline field values and references.
Converts pythonic types - eg None, True, int - to binary data.
Encoded inline and not stored directly in an object table. Do not contain
references.
"""






#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Python library for reading/writing Scratch project (.sb) and sprite files.

USAGE:
    You'll probably just want to use the provided ScratchProjectFile and 
    ScratchSpriteFile classes. Pass them the path to the file and use their 
    provided .save() methods. Access them using:
        from kurt.files import *
    
    Most of the objects you're interested in, like Stage and Sprite, inherit 
    from UserObject. You can use .fields.keys() to see the available fields on 
    one of these objects.
    
    FixedObjects have a .value property to access their value. Inline objects, 
    such as int and bool, are converted to their Pythonic counterparts. Array 
    and Dictionary are converted to list and dict.
    
    Tested with Python 2.6. Works with Scratch 1.4; not tested with earlier 
    versions, but probably works.
    
    Scratch is created by the Lifelong Kindergarten Group at the MIT Media Lab. 
    See their website: http://scratch.mit.edu/
"""

from kurt.objtable import *
from kurt.files import *
from kurt.scripts import *

try:
    from kurt.scratchblocks import parse_block_plugin
except ImportError:
    print "WARNING: parser not available, requires PLY"

__version__ = '1.4.5'



#coding=utf8

# Copyright © 2012 Tim Radvan
#
# This file is part of Kurt.
#
# Kurt is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Classes for manipulating scripts.
    Script - a stack; contains a list of blocks.
    Block - a single block.
"""

from fixed_objects import Symbol, Color, Point
from user_objects import BaseMorph

from pprint import pformat



class Block(object):
    """A single block.
    Arguments:
        script - the parent script that this block belongs to
        command - names the command this block performs (see BlockType.command)
        args - list of arguments for each of the block's inserts (see
               BlockType.defaults)
    Attributes:
        type - BlockType instance (found to match self.command)

    Methods:
        to_block_plugin() — returns the block in scratchblocks format.
    """
    def __init__(self, command_or_type=None, *args):
        self.script = None
        self._comment = None

        if not command_or_type:
            print 'OBSOLETE?'

        self.type = None
        if isinstance(command_or_type, BlockType):
            self.type = command_or_type
        else:
            command = command_or_type
            if isinstance(command, Symbol):
                command = command.value

            if command == 'scratchComment':
                raise ValueError, "Use ScratchComment class instead"

            if command in blocks_by_cmd:
                poss_types = blocks_by_cmd[command]
                if command == "changeVariable":
                    if len(args) < 3:
                        raise ValueError, \
                            "Invalid arguments to 'changeVariable' block"

                    for type in poss_types:
                        if type.defaults[1] == args[1]:
                            self.type = type
                            break
                    else:
                        raise ValueError, "Invalid argument " + repr(args[1]) \
                            + " to changeVariable block"

                elif command == "EventHatMorph":
                    if args[0] == "Scratch-StartClicked": # when gf clicked
                        for type in poss_types:
                            if type.defaults == ["Scratch-StartClicked"]:
                                self.type = type
                                break
                    else: # broadcast
                        for type in poss_types:
                            if type.defaults == []:
                                self.type = type
                                break
                else:
                    self.type = poss_types[0]

            else:
                print "WARNING: unknown block type %r" % command
                self.type = BlockType(command, command)

        if self.type:
            self.args = self.type.defaults[:]
        else:
            self.args = []

        for i in xrange(len(args)):
            if i < len(self.args):
                self.args[i] = args[i]
            else:
                self.args.append(args[i])

    @classmethod
    def from_array(cls, array):
        orig = array

        array = list(array)
        command = array.pop(0)

        args = []
        for arg in array:
            if isinstance(arg, list):
                if len(arg) == 0:
                    arg = Block.from_array([''])
                elif isinstance(arg[0], Symbol):
                    arg = Block.from_array(arg)
                else:
                    arg = [Block.from_array(block) for block in arg]
            args.append(arg)

        x = cls(command, *args)
        return x

    def to_array(self):
        array = []
        if self.command:
            array.append(Symbol(self.command))
        else:
            array.append('')

        for arg in self.args:
            if isinstance(arg, Block):
                array.append(arg.to_array())
            elif isinstance(arg, list):
                array.append([block.to_array() for block in arg])
            else:
                array.append(arg)
        return array

    def set_script(self, script):
        if self.script and self.script is not script:
            if self in self.script:
                self.script.remove(self)

        for arg in self.args:
            if isinstance(arg, Block):
                arg.set_script(script)
            elif isinstance(arg, list):
                for block in arg:
                    block.set_script(script)

        self.script = script

    def __eq__(self, other):
        return (
            isinstance(other, Block) and
            self.type == other.type and
            self.args == other.args
        )

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        string = "Block(%s," % repr(self.command)
        for arg in self.args:
            if isinstance(arg, Block):
                string = string.rstrip("\n")
                string += "\n\t" + repr(arg).replace("\n", "\n\t") + ",\n"
            elif isinstance(arg, list):
                if string.endswith("\n"):
                    string += "\t"
                else:
                    string += " "
                string += "[\n"
                for block in arg:
                    string += "\t\t" + repr(block).replace("\n", "\n\t\t")
                    string += ",\n"
                string += "\t], "
            else:
                string += repr(arg) + ", "
        string = string.rstrip(" ")
        string = string.rstrip(",")
        return string + ")"

    @property
    def name(self):
        print "WARNING: Block.name is deprecated -- use Block.command instead"
        # DEPRECATED -- TODO: remove
        return self.command

    @property
    def command(self):
        if self.type:
            return self.type.command
        return ""

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, comment):
        if comment and not isinstance(comment, Comment):
            comment = Comment(comment=comment, anchor=self)
        self._comment = comment
        if comment and not comment.anchor is self:
            comment.anchor = self

    def add_comment(self, comment):
        if not self.comment or isinstance(comment, Comment):
            self.comment = comment
        elif comment:
            self.comment.comment += "\n" + comment

    def to_block_plugin(self):
        arguments = self.args[:]

        def get_insert(value, insert_type=None):
            insert_fmt = None
            if insert_type:
                insert_fmt = block_plugin_inserts[insert_type]

            if isinstance(value, Block):
                block = value
                if block.command in ("readVariable", "contentsOfList:"):
                    value = block.args[0]
                    insert_fmt = "(%s)"
                else:
                    value = block.to_block_plugin()
                    if insert_type in ('%n', '%d', '%s', '%m', '%e', '%l',
                                          '%S', '%y', '%i'):
                        insert_fmt = "(%s)"
                    elif insert_type == "%b":
                        if (block.type and block.type.flag != 'b'):
                            # some booleans have to be encoded as reporters
                            insert_fmt = "(%s)"


                if not insert_fmt: insert_fmt = "(%s)"

            elif value == Symbol("mouse"):
                value = "mouse-pointer"
                if not insert_fmt: insert_fmt = block_plugin_inserts["%m"]

            elif value in (Symbol("last"), Symbol("all"), Symbol("any")):
                value = value.value
                if not insert_fmt: insert_fmt = block_plugin_inserts["%i"]

            elif isinstance(value, BaseMorph):
                value = value.name
                if not insert_fmt: insert_fmt = block_plugin_inserts["%m"]

            elif isinstance(value, Color):
                value = value.hexcode()
                if not insert_fmt: insert_fmt = block_plugin_inserts["%c"]

            elif isinstance(value, str) or isinstance(value, unicode):
                if not insert_fmt: insert_fmt = block_plugin_inserts["%s"]

            elif value is None or value is False:
                value = ""
                if not insert_fmt: insert_fmt = "[%s]"

            else:
                value = unicode(value)
                if not insert_fmt: insert_fmt = "[%s]"

            return insert_fmt % value


        block_type = self.type

        if self.command == "changeVariable":
            change = arguments.pop(1)
            if change.value == "setVar:to:":
                text = "set %v to %s"
            elif change.value == "changeVar:by:":
                text = "change %v by %n"
            block_type = BlockType(self.command, text)

        if not block_type:
            string = self.command
            for arg in self.args:
                arg = get_insert(arg)
                string += " " + arg

        else:
            if self.command == "MouseClickEventHatMorph":
                try:
                    morph_name = self.script.morph.name
                except AttributeError:
                    morph_name = "sprite"
                arguments[0] = morph_name

            #elif self.command == "EventHatMorph":
            #    if arguments[0] != "Scratch-StartClicked":
            #        block_type = BlockType(self.command, "when I receive %e")

            if self.command in ("not", "="): # fix weird blockplugin bug
                block_type = block_type.copy()
                block_type.text = block_type.text.replace(" ", "")

            string = ""
            for part in block_type.parts:
                if BlockType.INSERT_RE.match(part):
                    value = ""
                    if arguments:
                        value = arguments.pop(0)

                    string += get_insert(value, insert_type=part)
                else:
                    string += part

        if self.comment:
            string += " " + self.comment.to_block_plugin()

        if block_type and block_type.flag == "c":
            blocks = []
            if arguments:
                blocks = arguments.pop(0)
            if blocks:
                for block in blocks:
                    block_str = block.to_block_plugin()
                    string += "\n\t" + block_str.replace("\n", "\n\t")

            if self.command == 'doIfElse':
                string += '\nelse'

                blocks = []
                if arguments:
                    blocks = arguments.pop(0)
                if blocks:
                    for block in blocks:
                        block_str = block.to_block_plugin()
                        string += "\n\t" + block_str.replace("\n", "\n\t")

            string += "\nend"

        return string

    def to_block_list(self):
        yield self
        for arg in self.args:
            if isinstance(arg, Block):
                for block in arg.to_block_list():
                    yield block
            elif isinstance(arg, list):
                for block in arg:
                    for b in block.to_block_list():
                        yield b

    def replace_sprite_refs(self, lookup_sprite_named):
        """Replace all the kurt.scratchblocks.SpriteRef objects with Sprite
        objects with the corresponding name."""
        for i in range(len(self.args)):
            arg = self.args[i]
            if isinstance(arg, SpriteRef):
                self.args[i] = lookup_sprite_named(arg.name)
            elif isinstance(arg, Block):
                arg.replace_sprite_refs(lookup_sprite_named)
            elif isinstance(arg, list):
                for block in arg:
                    block.replace_sprite_refs(lookup_sprite_named)

    def find_undefined_variables(self, stage):
        """Make sure all the readVariable blocks are valid.
        Replace them with contentsOfList: if necessary"""
        if self.type.command == "readVariable":
            assert self.args
            var = self.args[0]
            morph = self.script.morph

            if var in morph.variables or var in stage.variables:
                return # we're good
            elif var in morph.lists or var in stage.lists:
                self.type = blocks_by_cmd["contentsOfList:"][0]
                return # Fixed!
            else:
                yield var
        else:
            for arg in self.args:
                if isinstance(arg, Block):
                    for var in arg.find_undefined_variables(stage):
                        yield var
                elif isinstance(arg, list):
                    for block in arg:
                        for var in block.find_undefined_variables(stage):
                            yield var


class Script(object):
    """A single stack of blocks.

    The first block self.blocks[0] is usually a “when” block, eg. an
    EventHatMorph.

    Scripts implement the `list` interface, so can be indexed directly; eg
    `script[0]` is the same as `script.blocks[0]`. All other methods like
    `append` also work.

    Arguments/attributes:
        morph - ScriptableScratchMorph (Stage or Sprite) instance that this
                script belongs to
        pos - (x, y) position of script in the “blocks bin”, the script pane in
              the Scratch interface.
        blocks - list of blocks.

    Methods:
        to_block_plugin() — returns the script in scratchblocks format.
    """
    def __init__(self, pos=(0,0), blocks=None):
        self.morph = None
        if blocks is None: blocks = []
        self.pos = pos
        self.blocks = blocks

        for block in blocks:
            if isinstance(block, Block): # Might still be Refs!
                block.set_script(self)

    @classmethod
    def from_array(cls, morph, array):
        (pos, blocks) = array

        if len(blocks) == 1:
            block = blocks[0]
            if block:
                command = block[0]
                if ( isinstance(command, Symbol) and
                     command.value == 'scratchComment' ):
                    return Comment.from_array(morph, pos, block)

        script = cls(pos, [Block.from_array(block) for block in blocks])
        script.morph = morph
        for block in script.blocks:
            if isinstance(block, Block):
                block.set_script(script)
        return script

    def to_array(self):
        return (self.pos, [block.to_array() for block in self.blocks])

    def __eq__(self, other):
        return (
            isinstance(other, Script) and
            self.blocks == other.blocks
        )

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        string = "Script(%s, [\n" % repr(self.pos)
        for block in self.blocks:
            string += "\t" + repr(block).replace("\n", "\n\t") + ",\n"
        string = string.rstrip()
        string = string.rstrip(",")
        return string + "])"

    def to_block_plugin(self):
        """Returns the script in scratchblocks format."""
        string = ""
        for block in self.blocks:
            string += block.to_block_plugin() + "\n"
        return string

    def to_block_list(self):
        for block in self.blocks:
            for b in block.to_block_list():
                yield b

    def replace_sprite_refs(self, lookup_sprite_named):
        """Replace all the kurt.scratchblocks.SpriteRef objects with Sprite
        objects with the corresponding name."""
        for block in self.blocks:
            block.replace_sprite_refs(lookup_sprite_named)

    def find_undefined_variables(self, stage):
        assert self.morph is not None
        for block in self.blocks:
            assert block.script == self
            for var in block.find_undefined_variables(stage):
                yield var


    # Pretend to be a list #

    def __iter__(self):
        return iter(self.blocks)

    def __getattr__(self, name):
        return getattr(self.blocks, name)

    def append(self, block):
        self.blocks.append(block)
        block.set_script(self)

    def insert(self, index, block):
        self.blocks.insert(index, block)
        block.set_script(self)

    def __getitem__(self, index):
        return self.blocks[index]

    def __setitem__(self, index, value):
        self.blocks[index] = value

    def __delitem__(self, index):
        del self.blocks[index]

    def __len__(self):
        return len(self.blocks)



class ScriptCollection(list):
    def __init__(self, scripts=None):
        if scripts is None: scripts = []
        self += scripts

    def __repr__(self):
        return pformat(list(self))



# circular imports are stupid
from blockspecs import BlockType
from blockspecs import blocks_by_cmd, block_plugin_inserts, find_block



class Comment(Script):
    type = BlockType("scratchComment", "// %s", defaults = ["", True, 112])

    def __init__(self, pos=None, comment="", showing=True, width=112, anchor=None):
        self.morph = None
        self.blocks = []
        if pos is None:
            pos = Point(0, 0)
        self.pos = pos

        self.comment = comment
        self.showing = showing
        self.width = width
        self._anchor = anchor

    def __len__(self):
        return True

    @classmethod
    def from_array(cls, morph, pos, block):
        if block[0] == Symbol('scratchComment'):
            block.pop(0)
        comment = cls(pos, *block)
        comment.morph = morph
        return comment

    def to_array(self, blocks_by_id):
        array = [Symbol('scratchComment')]
        array += [self.comment, self.showing, self.width]
        if self.anchor:
            for i in xrange(len(blocks_by_id)):
                if blocks_by_id[i] is self.anchor:
                    array.append(i + 1)
                    break
        return (self.pos, [array])

    def __repr__(self):
        return 'Comment' + repr((
            self.pos, self.comment, self.showing, self.width,
        ))

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self, block):
        self._anchor = block
        if block:
            if not self.pos and block.script:
                (x, y) = block.script.pos
                self.pos = (x + 200, y)

            if not block.comment is self:
                block.comment = self

    def attach_scripts(self, blocks_by_id):
        """Attach the comment to the right block from the given scripts."""
        if self.anchor:
            if not isinstance(self.anchor, Block):
                index = self.anchor
                self.anchor = blocks_by_id[index - 1]
                self.anchor.comment = self

    def to_block_plugin(self):
        return '// ' + self.comment.replace('\r', '\n').replace('\n', ' ')



class SpriteRef(object):
    """Used by `kurt.scratchblocks.parser` so that compile() doesn't have to
    build all the sprites before parsing all the scripts."""
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'SpriteRef(%s)' % self.name









#coding=utf8

# Copyright © 2012 Tim Radvan
# 
# This file is part of Kurt.
# 
# Kurt is free software: you can redistribute it and/or modify it under the 
# terms of the GNU Lesser General Public License as published by the Free 
# Software Foundation, either version 3 of the License, or (at your option) any 
# later version.
# 
# Kurt is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along 
# with Kurt. If not, see <http://www.gnu.org/licenses/>.

"""Classes for saving/loading Scratch files.
Construct them by passing them a path.

    ScratchProjectFile - .sb
    ScratchSpriteFile - .sprite
"""

from construct import Container, Struct, Bytes, Rename
from construct.text import Literal
from objtable import ObjTable, InfoTable
from fixed_objects import Form
from user_objects import Stage, Image

import os.path



class BinaryFile(object):
    """File(path)
    Implements a basic file with save() function.
    Subclasses override _load() and _save()."""
    
    EXTENSION = None
    
    def __init__(self, path=None, load=True):
        """Loads a file.
        @param path: the path passed to open().
        """
        self.path = None
        if path:
            if not path.lower().endswith("."+self.EXTENSION.lower()):
                path += "."+self.EXTENSION
            self.path = path
        
            if load:
                self.load()
    
    @property
    def name(self):
        (folder, name) = os.path.split(self.path)
        trimmed_name = ''.join(name.split('.')[:-1]) # Trim extension
        if trimmed_name:
            name = trimmed_name
        return name
    
    def load(self):
        """Reload the file from disk, replacing any changes in memory.
        """
        if not self.path:
            raise ValueError, "filepath not set."
        
        f = open(self.path, "rb")
        bytes = f.read()
        f.close()
        self._load(bytes)
    
    def _load(self, bytes):
        """Subclasses must override this method.
        Set the attributes of this file from the given contents.
        @param bytes: str containing the file contents read from disk.
        """
        raise NotImplementedError()
    
    def save(self, path=None):
        """Save the file to disk.
        @param path: (optional) set new destination path. Future saves will go 
                     to the new location.
        """
        if path and not path.lower().endswith("."+self.EXTENSION.lower()):
            path += "."+self.EXTENSION
        if path:
            self.path = path
        if not self.path:
            raise ValueError, "filepath not set."
        
        bytes = self._save()
        if not bytes:
            print "Can't write zero bytes to file, aborting"
            return
        
        f = open(self.path, 'wb')
        f.write(bytes)
        f.flush()
        f.close()
    
    def _save(self):
        """Subclasses must override this method.
        @return: str containing the bytes to be saved to disk.
        """
        raise NotImplementedError()
    
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.path))



class ScratchProjectFile(BinaryFile):
    """A Scratch Project file.
    @param path: path to .sb file.
    
    Attributes:
        info - a Dictionary containing project info (author, notes, thumbnail)
        stage - the stage. Contains contents, including sprites and media.
    """
    
    EXTENSION = "sb"
    
    DEFAULT_COMMENT = "Made with Kurt \nhttp://github.com/blob8108/kurt"
    DEFAULT_INFO = {
        "comment": DEFAULT_COMMENT,
        "scratch-version": '1.4 of 30-Jun-09',
        "language": "en",
        "author": u"",
        "isHosting": False,
        "platform": "", 
        "os-version": "",
        "thumbnail": None,
        "history": "",
    }
    
    _construct = Struct("scratch_file",
        Literal("ScratchV02"),
        Rename("info", InfoTable),
        Rename("stage", ObjTable),
    )
    
    def __init__(self, *args, **kwargs):
        self.info = self.DEFAULT_INFO.copy()
        BinaryFile.__init__(self, *args, **kwargs)

    def _load(self, bytes):
        project = self._construct.parse(bytes)
        self.info.update(project.info)
        
        if self.info["thumbnail"] and isinstance(self.info["thumbnail"], Form):
            self.info["thumbnail"] = Image(
                name = self.name + " thumbnail",
                form = self.info["thumbnail"],
            )
        
        self.info["comment"] = self.info["comment"].replace("\r", "\n")
        
        #self.info.__doc__ = InfoTable.__doc__
        self.stage = project.stage
    
    def _save(self):        
        info = self.info.copy()
        if info["thumbnail"]:
            info["thumbnail"] = info["thumbnail"].form
        info["comment"] = info["comment"].replace("\n", "\r")
        
        self.stage.normalize()
        
        project = Container(
            info = info,
            stage = self.stage,
        )
        return self._construct.build(project)
    
    @classmethod
    def new(cls, path=None):
        """Returns a new, empty project.
        Optional path argument.
        Will not write to disk until you .save()
        """
        project = cls(path, load=False)
        project.stage = Stage()
        #project.path = path # do this now so project doesn't attempt
        #                    # to .load() itself
        return project
    
    def __getattr__(self, name):
        if name in self.DEFAULT_INFO:
            return self.info[name]
    
    def __setattr__(self, name, value):
        if name in self.DEFAULT_INFO:
            self.info[name] = value
        else:
            object.__setattr__(self, name, value)
    
    @property
    def sprites(self):
        return self.stage.sprites
        
    @sprites.setter
    def sprites(self, value):
        self.stage.sprites = value
    
    def get_sprite(self, name):
        for sprite in self.sprites:
            if sprite.name == name:
                return sprite


class ScratchSpriteFile(BinaryFile):
    """A Scratch sprite file.
    @param path: path to .sprite file.
    
    Attributes:
        stage - the root object of the file (Sprite files actually contain a 
                serialised Stage)
        sprite - convenience property for accessing the first (only) sprite in 
                 the file.
    """

    EXTENSION = "sprite"
    
    def _load(self, bytes):
        self.stage = ObjTable.parse(bytes)
    
    def _save(self):
        return ObjTable.build(self.stage)
    
    @property
    def sprite(self):
        return self.stage.submorphs[0]



__all__ = ['ScratchProjectFile', 'ScratchSpriteFile']


